/*
 * Control bar
 *
 * Copyright (C) 2000-2002, Nam SungHyun and various contributors
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#include "config.h"
#include <time.h>	/* time(), localtime() */
#include <fcntl.h>	/* open() */

#include "pcMain.h"
#include "pcTerm.h"
#include <gdk/gdkx.h>

/* Global variables/functions {{{1 */

guint32 MaxCaptureSize; /* maximum capture file size */
char *CapturePath;
char *CaptureFile;
char *LogFile;

/* Local variables {{{1 */

static GtkWidget *BaudFrame = NULL, *BaudLabel = NULL;
static GtkWidget *EmulFrame = NULL, *EmulLabel = NULL;
static GtkWidget *ProtFrame = NULL, *ProtLabel = NULL;

/* pcMain.h emulate enum  ƾ Ѵ. */
const char *EmulateNames[] = { "ANSI", "VT100" };

/* ControlBaudrateChange() {{{1 */
void
ControlBaudrateChange(guint baud)
{
    char buf[9];

    if (BaudLabel)
    {
	g_snprintf(buf, sizeof(buf), " %6d ", baud);
	gtk_label_set_text(GTK_LABEL(BaudLabel), buf);
    }
}

/* ControlProtocolChange() {{{1 */
void
ControlProtocolChange(void)
{
    char buf[9];

    if (ProtLabel)
    {
	g_snprintf(buf, sizeof(buf), " %6s ", TRxProtocolNames[TRxProtocol]);
	gtk_label_set_text(GTK_LABEL(ProtLabel), buf);
    }
}

/* ControlEmulateChange() {{{1 */
void
ControlEmulateChange(void)
{
    char buf[8];

    if (EmulLabel)
    {
	g_snprintf(buf, sizeof(buf), " %5s ", EmulateNames[EmulateMode]);
	gtk_label_set_text(GTK_LABEL(EmulLabel), buf);
    }
}

/* CreateControlLabel() {{{1 */
static GtkWidget *
CreateControlLabel(char *label, GtkWidget **labelWin)
{
    GtkWidget *frame;
    char *p;

    frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
    gtk_box_pack_end(GTK_BOX(CtrlBox), frame, FALSE, FALSE, 0);

    p = PC_IconvStr(label, strlen(label), 256);
    *labelWin = gtk_label_new(p);
    if (p != label)
	g_free(p);

    gtk_container_add(GTK_CONTAINER(frame), *labelWin);
    gtk_widget_show_all(frame);
    return frame;
}

/* ControlLabelBaudCtrl() {{{1 */
int
ControlLabelBaudCtrl(gboolean create)
{
    char buf[10];

    if (create)
    {
	if (!BaudFrame)
	{
	    g_snprintf(buf, sizeof(buf), " %6d ", ModemGetSpeed());
	    BaudFrame = CreateControlLabel(buf, &BaudLabel);
	    return OK;
	}
    }
    else
    {
	if (BaudFrame)
	{
	    gtk_widget_destroy(BaudFrame);
	    BaudLabel = BaudFrame = NULL;
	    return OK;
	}
    }
    return ERROR;
}

/* ControlLabelEmulCtrl() {{{1 */
int
ControlLabelEmulCtrl(gboolean create)
{
    char buf[10];

    if (create)
    {
	if (!EmulFrame)
	{
	    g_snprintf(buf, sizeof(buf), " %5s ", EmulateNames[EmulateMode]);
	    EmulFrame = CreateControlLabel(buf, &EmulLabel);
	    return OK;
	}
    }
    else
    {
	if (EmulFrame)
	{
	    gtk_widget_destroy(EmulFrame);
	    EmulLabel = EmulFrame = NULL;
	    return OK;
	}
    }
    return ERROR;
}

/* ControlLabelProtCtrl() {{{1 */
int
ControlLabelProtCtrl(gboolean create)
{
    char buf[10];

    if (create)
    {
	if (!ProtFrame)
	{
	    g_snprintf(buf, sizeof(buf), " %5s ",
		       TRxProtocolNames[TRxProtocol]);
	    ProtFrame = CreateControlLabel(buf, &ProtLabel);
	    return OK;
	}
    }
    else
    {
	if (ProtFrame)
	{
	    gtk_widget_destroy(ProtFrame);
	    ProtLabel = ProtFrame = NULL;
	    return OK;
	}
    }
    return ERROR;
}
/* }}} */

/****************************************************************************
		     C O N T R O L   L I N E   S T A T U S
****************************************************************************/
#include "pixmaps/yellow.xpm"
#include "pixmaps/red.xpm"
#include "pixmaps/green.xpm"

#define LED_ON_INTERVAL 100

static GtkWidget *ctrlLineSts = NULL;
static GtkWidget *ctrlRxBox, *ctrlTxBox;
static GtkWidget *ctrlRxOffWid, *ctrlTxOffWid, *ctrlTxWid, *ctrlRxWid;
static int ctrlRxTid, ctrlTxTid;
static GMutex *ctrlMutex = NULL;

/* ControlLabelLineStsCtrl() {{{1 */
int
ControlLabelLineStsCtrl(gboolean create)
{
    GtkWidget *label, *box;
    GdkPixmap *pixYellow, *pixRed, *pixGreen;
    GdkBitmap *maskYellow, *maskRed, *maskGreen;

    if (ctrlMutex == NULL)
	ctrlMutex = g_mutex_new();

    if (ctrlLineSts)
    {
	if (create == FALSE)
	{
	    g_mutex_lock(ctrlMutex);
	    gtk_widget_destroy(ctrlLineSts);
	    ctrlLineSts = NULL;
	    ctrlRxTid = -1;
	    ctrlTxTid = -1;
	    g_mutex_unlock(ctrlMutex);
	    return OK;
	}
	return ERROR;
    }

    ctrlRxTid = -1;
    ctrlTxTid = -1;

    pixYellow = gdk_pixmap_create_from_xpm_d(GDK_ROOT_PARENT(), &maskYellow,
					    NULL, yellow_xpm);
    pixRed = gdk_pixmap_create_from_xpm_d(GDK_ROOT_PARENT(), &maskRed,
					  NULL, red_xpm);
    pixGreen = gdk_pixmap_create_from_xpm_d(GDK_ROOT_PARENT(), &maskGreen,
					    NULL, green_xpm);
    if (pixYellow == NULL || pixRed == NULL || pixGreen == NULL)
    {
	g_warning("Cannot create pixmaps\n");
	return ERROR;
    }
    ctrlRxWid = gtk_pixmap_new(pixGreen, maskGreen);
    ctrlTxWid = gtk_pixmap_new(pixRed, maskRed);
    gtk_widget_show(ctrlRxWid);
    gtk_widget_show(ctrlTxWid);

    ctrlLineSts = gtk_frame_new(0);
    gtk_frame_set_shadow_type(GTK_FRAME(ctrlLineSts), GTK_SHADOW_IN);
    gtk_box_pack_end(GTK_BOX(CtrlBox), ctrlLineSts, FALSE, FALSE, 0);

    box = gtk_hbox_new(0, 0);
    gtk_container_add(GTK_CONTAINER(ctrlLineSts), box);

    ctrlRxBox = gtk_hbox_new(0, 0);
    gtk_box_pack_end(GTK_BOX(box), ctrlRxBox, FALSE, FALSE, 0);

    ctrlRxOffWid = gtk_pixmap_new(pixYellow, maskYellow);
    gtk_box_pack_start(GTK_BOX(ctrlRxBox), ctrlRxOffWid, FALSE, FALSE, 5);
    label = gtk_label_new("R");
    gtk_box_pack_end(GTK_BOX(ctrlRxBox), label, FALSE, FALSE, 5);

    ctrlTxBox = gtk_hbox_new(0, 0);
    gtk_box_pack_end(GTK_BOX(box), ctrlTxBox, FALSE, FALSE, 0);

    ctrlTxOffWid = gtk_pixmap_new(pixYellow, maskYellow);
    gtk_box_pack_start(GTK_BOX(ctrlTxBox), ctrlTxOffWid, FALSE, FALSE, 5);
    label = gtk_label_new("T");
    gtk_box_pack_end(GTK_BOX(ctrlTxBox), label, FALSE, FALSE, 5);

    gtk_widget_show_all(ctrlLineSts);
    return OK;
}

/* ctrlRxToggleButtonOff() {{{1 */
static int
ctrlRxToggleButtonOff(void)
{
    g_mutex_lock(ctrlMutex);

    if (ctrlLineSts)
    {
	gtk_widget_ref(ctrlRxWid);
	gtk_container_remove(GTK_CONTAINER(ctrlRxBox), ctrlRxWid);
	gtk_box_pack_start(GTK_BOX(ctrlRxBox), ctrlRxOffWid, FALSE,
			   FALSE, 5);
	ctrlRxTid = -1;
    }

    g_mutex_unlock(ctrlMutex);
    return FALSE;
}

/* ctrlTxToggleButtonOff() {{{1 */
static int
ctrlTxToggleButtonOff(void)
{
    g_mutex_lock(ctrlMutex);

    if (ctrlLineSts)
    {
	gtk_widget_ref(ctrlTxWid);
	gtk_container_remove(GTK_CONTAINER(ctrlTxBox), ctrlTxWid);
	gtk_box_pack_start(GTK_BOX(ctrlTxBox), ctrlTxOffWid, FALSE,
			   FALSE, 5);
	ctrlTxTid = -1;
    }

    g_mutex_unlock(ctrlMutex);
    return FALSE;
}

/* ControlLabelLineStsCtrlToggle() {{{1 */
void
ControlLabelLineStsToggle(int id)
{
    if (ctrlMutex == NULL)
	return;

    g_mutex_lock(ctrlMutex);

    if (ctrlLineSts)
    {
	if (id == 'r')
	{
	    if (ctrlRxTid < 0)
	    {
		gtk_widget_ref(ctrlRxOffWid);
		gtk_container_remove(GTK_CONTAINER(ctrlRxBox), ctrlRxOffWid);
		gtk_box_pack_start(GTK_BOX(ctrlRxBox), ctrlRxWid, FALSE,
				   FALSE, 5);
		ctrlRxTid = gtk_timeout_add(LED_ON_INTERVAL,
					    (GtkFunction) ctrlRxToggleButtonOff,
					    NULL);
	    }
	}
	else if (id == 't')
	{
	    if (ctrlTxTid < 0)
	    {
		gtk_widget_ref(ctrlTxOffWid);
		gtk_container_remove(GTK_CONTAINER(ctrlTxBox), ctrlTxOffWid);
		gtk_box_pack_start(GTK_BOX(ctrlTxBox), ctrlTxWid, FALSE,
				   FALSE, 5);

		ctrlTxTid = gtk_timeout_add(LED_ON_INTERVAL,
					    (GtkFunction) ctrlTxToggleButtonOff,
					    NULL);
	    }
	}
    }
    g_mutex_unlock(ctrlMutex);
}
/* }}} */

/****************************************************************************
			    C O N T R O L   M E N U
****************************************************************************/
/* CtrlMenuListType {{{1 */
typedef struct {
    gboolean global;
    char *title;
    GtkWidget *menu;
    GtkWidget *menuBar;
    int entryNum;
} CtrlMenuListType;

static GSList *CtrlMenuList = NULL;
static gboolean CtrlMenuRemoved = TRUE;

/* RunControlMenuString() {{{1 */
static void
RunControlMenuString(GtkWidget *w, char *string)
{
    Term->remoteWrite(string, strlen(string));
    SoundPlay(SoundClickFile);
    w = NULL;
}

/* RunControlMenuScript() {{{1 */
static void
RunControlMenuScript(GtkWidget *w, char *action)
{
    ScriptRunFromString(action);
    SoundPlay(SoundClickFile);
    w = NULL;
}

/* RunControlMenuScriptFile() {{{1 */
static void
RunControlMenuScriptFile(GtkWidget *w, char *action)
{
    ScriptRunFromFile(action);
    SoundPlay(SoundClickFile);
    w = NULL;
}

/* CtrlMenuAddEntry() {{{1 */
static void
CtrlMenuAddEntry(GtkWidget *menu, const char *name, GtkSignalFunc func,
		 gpointer data)
{
    GtkWidget *menuItem;

    menuItem = gtk_menu_item_new_with_label(name);
    gtk_signal_connect(GTK_OBJECT(menuItem), "activate", func, data);
    gtk_menu_append(GTK_MENU(menu), menuItem);
    gtk_widget_show(menuItem);
}

/* CreateCtrlMenu() {{{1 */
static void
CreateCtrlMenu(CtrlMenuListType *cml, const char *title,
	       const char *name, GtkSignalFunc func, gpointer data)
{
    GtkWidget *menuBar, *menu, *menuItem, *hbox, *arrow, *label;

    menuBar = gtk_menu_bar_new();
    gtk_box_pack_start(GTK_BOX(CtrlBox), menuBar, FALSE, TRUE, 0);
    gtk_container_set_border_width(GTK_CONTAINER(menuBar), 0);
    gtk_widget_show(menuBar);

    menu = gtk_menu_new();

    CtrlMenuAddEntry(menu, name, func, data);

    menuItem = gtk_menu_item_new();
    hbox = gtk_hbox_new(FALSE, 1);
    arrow = gtk_arrow_new(GTK_ARROW_DOWN, GTK_SHADOW_OUT);
    label = gtk_label_new(title);
    gtk_box_pack_start(GTK_BOX(hbox), arrow, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 4);
    gtk_widget_show(arrow);
    gtk_widget_show(label);
    gtk_widget_show(hbox);
    gtk_container_add(GTK_CONTAINER(menuItem), hbox);
    gtk_container_set_border_width(GTK_CONTAINER(menuItem), 0);
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuItem), menu);
    gtk_menu_bar_append(GTK_MENU_BAR(menuBar), menuItem);
    gtk_widget_show(menuItem);

    cml->menu = menu;
    cml->menuBar = menuBar;
}

/* ControlMenuExit() {{{1 */
void
ControlMenuExit(void)
{
    GSList *list;
    CtrlMenuListType *cml;

    if (!CtrlMenuRemoved)
    {
	CtrlMenuRemoved = TRUE;
	DisconnectFuncList = g_slist_remove(DisconnectFuncList,
					    ControlMenuExit);
    }

    for (list = CtrlMenuList; list;)
    {
	cml = list->data;
	list = list->next;
	if (!cml->global)
	{
	    gtk_widget_destroy(cml->menuBar);
	    CtrlMenuList = g_slist_remove(CtrlMenuList, cml);
	    g_free(cml->title);
	    g_free(cml);
	}
    }
}

/* ControlBarAddMenu() {{{1 */
void
ControlBarAddMenu(gboolean globalf, const char *title, const char *name,
		  const char *type, const char *action)
{
    GSList *list;
    CtrlMenuListType *cml = NULL;
    GtkSignalFunc func;
    gpointer data;

    for (list = CtrlMenuList; list; list = list->next)
    {
	cml = list->data;
	if (!strcmp(title, cml->title))
	    break;
    }

    if (!strcasecmp(type, "string"))
	func = (GtkSignalFunc) RunControlMenuString;
    else if (!strcasecmp(type, "script"))
	func = (GtkSignalFunc) RunControlMenuScript;
    else if (!strcasecmp(type, "sfile"))
	func = (GtkSignalFunc) RunControlMenuScriptFile;
    else
	return;

    data = ConvertCtrlChar(action);

    if (!list)
    {
	cml = g_new(CtrlMenuListType, 1);
	cml->title = g_strdup(title);
	cml->entryNum = 1;
	cml->global = globalf;
	CreateCtrlMenu(cml, title, name, func, data);
	CtrlMenuList = g_slist_append(CtrlMenuList, cml);

	if (!cml->global && CtrlMenuRemoved)
	{
	    CtrlMenuRemoved = FALSE;
	    DisconnectFuncList = g_slist_append(DisconnectFuncList,
						ControlMenuExit);
	}
    }
    else
    {
	if (cml->global != globalf)
	    g_warning("You should use same token. %s added with %s",
		      title, cml->global ? "global" : "temporary");

	cml = list->data;
	cml->entryNum++;
	CtrlMenuAddEntry(cml->menu, name, func, data);
    }
}
/* }}} */

/****************************************************************************
			 C A P T U R E   C O N T R O L
****************************************************************************/
static int CaptureFD = -1;
static guint32 CurrCaptureSize = 0;
static GtkWidget *CaptureFrame = NULL;
static gchar *oldCapPath = NULL;

/* CaptureInputFilter() {{{1 */
int
CaptureInputFilter(const char *s, int len)
{
    register int i;

    /* download/upload  ̸ capture ƶ */
    if (CaptureFD != -1 && !TRxRunning)
    {
	if (CurrCaptureSize > MaxCaptureSize - len)
	{
	    if (MaxCaptureSize > CurrCaptureSize)
		len = MaxCaptureSize - CurrCaptureSize;
	    else
	    {
		g_warning("FIXME: yeah! it should be a BUG!");
		len = 1;
	    }
	}

	if (TermStripLineFeed)
	{
	    for (i = 0; i < len; i++)
	    {
		if (s[i] != '\r')
		{
		    write(CaptureFD, s + i, 1);
		    ++CurrCaptureSize;
		}
	    }
	}
	else
	{
	    write(CaptureFD, s, len);
	    CurrCaptureSize += len;
	}
	if (CurrCaptureSize >= MaxCaptureSize)
	    CaptureFinish();
    }
    return len;
}

/* CaptureFinish() {{{1 */
void
CaptureFinish(void)
{
    if (CaptureFD != -1)
    {
	if (!TermStripControlChar)
	    InputFilterList =
		g_slist_remove(InputFilterList, CaptureInputFilter);
	close(CaptureFD);
	CaptureFD = -1;
	CurrCaptureSize = 0UL;
	StatusShowMessage(_("Capture finished."));
	if (CaptureFrame)
	{
	    StatusRemoveFrame(CaptureFrame);
	    CaptureFrame = NULL;
	}
    }
}

/* CaptureFileSelectOk() {{{1 */
static void
CaptureFileSelectOk(GtkWidget *w, GtkWidget *win)
{
    const char *p, *filename;

    UNUSED(w);

    filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(win));
    if ((p = strrchr(filename, '/')) != NULL)
    {
	if (oldCapPath)
	    g_free(oldCapPath);
	oldCapPath = g_strdup(filename);
	oldCapPath[(int)(p - filename + 1)] = 0;
    }
    CaptureStart(filename, FALSE);
    gtk_widget_destroy(win);
}

/* CaptureFileQuery() {{{1 */
static void
CaptureFileQuery(void)
{
    GtkWidget *fs;
    char *file;

    fs = gtk_file_selection_new("Capture File");
    gtk_file_selection_hide_fileop_buttons(GTK_FILE_SELECTION(fs));
    file = g_strconcat(CapturePath, "/", NULL);
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(fs), file);
    g_free(file);
    gtk_window_set_position(GTK_WINDOW(fs), GTK_WIN_POS_MOUSE);
    gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fs)->ok_button),
		       "clicked", GTK_SIGNAL_FUNC(CaptureFileSelectOk), fs);
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(fs)->cancel_button),
			      "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
			      (GtkObject *) fs);
    gtk_widget_show(fs);
    if (oldCapPath)
	gtk_file_selection_set_filename(GTK_FILE_SELECTION(fs), oldCapPath);
}

/* CaptureStart() {{{1 */
int
CaptureStart(const char *filename, gboolean includeCurrBuf)
{
    int i;
    char buf[MAX_PATH];

    /* script  event ( auto-response)  capture ONϴ  
     * Լ  ҷ  ִ. ̹ captureϰ ִ 쿡 OK
     * return ش.
     */
    if (CaptureFD != -1)
	return OK;

    CurrCaptureSize = 0UL;

    if (!filename || !filename[0])
    {
	/* auto generated with date */
	time_t ti;
	struct tm *lt;

	ti = time(NULL);
	lt = localtime(&ti);
	g_snprintf(buf, sizeof(buf), "%s/%02d%02d%02d.cap",
		   CapturePath, lt->tm_year, lt->tm_mon + 1, lt->tm_mday);
	CaptureFD = open(buf, O_WRONLY | O_CREAT | O_APPEND, 0660);
    }
    else if (filename[0] == '?')
    {
	CaptureFileQuery();
	return OK;
    }
    else
    {
	if (filename[0] == '/')
	    strncpy(buf, filename, sizeof(buf) - 1);
	else
	    g_snprintf(buf, sizeof(buf), "%s/%s", CapturePath, filename);
	CaptureFD = open(buf, O_WRONLY | O_CREAT | O_APPEND, 0660);
    }

    if (CaptureFD == -1)
    {
	StatusShowMessage(_("Cannot open capture file '%s'."), buf);
	return ERROR;
    }
    else
    {
	if (!TermStripControlChar)
	    InputFilterList =
		g_slist_append(InputFilterList, CaptureInputFilter);

	StatusShowMessage(_("Capture starts to '%s'."), buf);

	if (includeCurrBuf)
	{
	    int maxLine;

	    maxLine = Term->visibleLineBegin;
	    if ((guint32) maxLine * Term->col > MaxCaptureSize)
		i = maxLine - MaxCaptureSize / Term->col + 1;
	    else
		i = 0;
	    for (; i < maxLine; i++)
	    {
		write(CaptureFD, Term->buf + i * Term->col, Term->col);
		write(CaptureFD, "\n", 1);
		CurrCaptureSize += Term->col + 1;
	    }
	}

	if (!CaptureFrame)
	    CaptureFrame = StatusAddFrameWithLabel(" LOG ");

	return OK;
    }
}

/* CaptureControl() {{{1 */
void
CaptureControl(void)
{
    if (CaptureFD != -1)
	CaptureFinish();
    else
	CaptureStart(CaptureFile, FALSE);
}
/* }}} */

/****************************************************************************
			 L O G G I N G   C O N T R O L
****************************************************************************/
static FILE *LogFP = NULL;

/* OpenLogFile() {{{1 */
void
OpenLogFile(const char *name, const char *info)
{
    char *fname;
    time_t tim;
    struct tm *lt;
    char buf[128];

    if (LogFile && *LogFile)
    {
	if (*LogFile == '/' || *LogFile == '~')
	    fname = g_strdup(ExpandPath(LogFile));
	else
	    fname = g_strconcat(ExpandPath(PC_RC_BASE), "/", LogFile, NULL);

	if ((LogFP = fopen(fname, "a")) == NULL)
	    StatusShowMessage(_("Cannot open log file '%s'"), fname);
	else
	{
	    tim = time(NULL);
	    lt = localtime(&tim);
	    strftime(buf, sizeof(buf), "%a %b %d %T %Z %Y", lt);
	    fprintf(LogFP, "===================================\n");
	    fprintf(LogFP, "%s: %s\nStart: %s\n", name, info, buf);
	    fflush(LogFP);
	}
	g_free(fname);
    }
}

/* CloseLogFile() {{{1 */
void
CloseLogFile(gint32 tick)
{
    if (LogFP)
    {
	if (tick < 0)
	    tick = 0;
	fprintf(LogFP, "Total conection time: %02d:%02d:%02d\n\n",
		tick / 3600, (tick % 3600) / 60, tick % 60);
	fclose(LogFP);
	LogFP = NULL;
    }
}

/* AppendLogData() {{{1 */
/* NOT USED!!!
void
AppendLogData(const gchar *format, ...)
{
   va_list args;
   gchar *buf;

   if (LogFP)
   {
      va_start(args, format);
      buf = g_strdup_vprintf(format, args);
      va_end(args);

      fprintf(LogFP, buf);
      fflush(LogFP);
      g_free(buf);
   }
}
*/
