/*
 * GTK+ ׿ ް  ȯ .
 *
 * Copyright (C) 2000-2002, Nam SungHyun and various contributors
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#include "config.h"
#include <signal.h> /* SIGSEGV */
#include <gdk_imlib.h> /* gdk_imlib_init() */

#include "pcMain.h"
#include "pcTerm.h" /* TermType */
#include "pcChat.h"

/* Global variables/functions {{{1 */

const char *MyName = NULL; /*  ̸ */

GtkWidget *MainWin = NULL;
GtkWidget *MainBox = NULL;
GtkWidget *TermBox = NULL;
GtkWidget *CtrlBox = NULL;
GtkWidget *StatBox = NULL;
GtkItemFactory *MenuFactory = NULL;
GtkWidget *ScriptBox = NULL;

guint PC_State = PC_IDLE_STATE;
gboolean UseModem = TRUE;
int ReadFD = -1;
guint32 LastOutTime = 0;
gboolean DisconnectAfterTransfer = FALSE;

gboolean ControlBarShow;
char *ToolBarType;
gboolean ToolBarShow;
gboolean UseHandleBox;

GSList *InputFilterList = NULL;
GSList *ConnectFuncList = NULL;
GSList *DisconnectFuncList = NULL;

char *IconvToStr;
char *IconvFromStr;

#ifdef _DEBUG_
gboolean OPrintFlag[MAX_OPRT_NUM];
#endif

/* Local variables {{{1 */

static GtkWidget *TB_OpenButton = NULL, *MenuOpenButton = NULL;
static GtkWidget *TB_CloseButton = NULL, *MenuCloseButton = NULL;
static GtkWidget *TB_UploadButton = NULL, *MenuUploadButton = NULL;
static GtkWidget *TB_DownloadButton = NULL, *MenuDownloadButton = NULL;
static GtkWidget *MenuStopTRxButton = NULL;
static GtkWidget *ToolBox = NULL;
static GtkWidget *HandleBox = NULL;
static GtkWidget *MenuOpenWin = NULL;

static char *MyFullName = NULL;

/* caughtSignal() {{{1 */
static RETSIGTYPE
caughtSignal(int sig)
{
    fprintf(stderr, "%s: caught signal %d\n", MyName, sig);

#ifdef _DEBUG_
    if (MyFullName)
    {
	char tmpFileName[] = "/tmp/.pcgdb.XXXXXX";
	char cmd[256];
	FILE *fp;

	mkstemp(tmpFileName);
	if ((fp = fopen(tmpFileName, "w")) != NULL)
	{
	    fprintf(fp,
		    "file %s\n"
		    "attach %d\n"
		    "set height 1000\n"
		    "bt full\n"
		    , MyFullName, getpid());
	    fclose(fp);
	    g_snprintf(cmd, sizeof(cmd), "xterm -sb -sl 1024 -e gdb -x %s", tmpFileName);
	    system(cmd);
	    unlink(tmpFileName);
	}
    }
#endif

    /* g_on_error_query(NULL); */
    gdk_exit(1);
    _exit(1);
}

/* PC_Hangup() {{{1 */
void
PC_Hangup(void)
{
    if (UseModem)
	ModemHangup();
    else
	TelnetDone();
}

/* DoExitClean() {{{1 */
static void
DoExitClean(void)
{
    SoundPlay(SoundExitFile);
    PC_Hangup();	/* hangup the modem or telnet if it was opened */
    TRxStop();	/* file ̸  */
    CaptureFinish();	/* close capture file if it was opened. */
    ChatExit();
    I_SelectExit();
    AutoResponseExit();
    ControlMenuExit();
    ScriptExit();
    ModemExit();
    TelnetExit();
    ConfigExit();
    TermExit();
    ChildSignalHandlerExit();
    if (InputFilterList)
	g_slist_free(InputFilterList);
    if (ConnectFuncList)
	g_slist_free(ConnectFuncList);
    if (DisconnectFuncList)
	g_slist_free(DisconnectFuncList);
    ScriptRunAnimDestroy();
    PC_IconvClose();
}

/* DoExit() {{{1 */
void
DoExit(int code)
{
    DoExitClean();
    gtk_main_quit();
    _exit(code);
}

/* aboutDialog() {{{1 */
static void
aboutDialog(void)
{
    GtkWidget *win, *label;
    static const char s[] =
	N_("This is a communication program which is free\n"
	   "with source code. Features include direct connection\n"
	   "via serial device, telnet connection, modem connection\n"
	   "script language support, capture to file, and more.\n"
	   "\n"
	   "It is based on GAU made by Chi-Deok Hwang. Many\n"
	   "thanks to him\n"
	   "\n"
	   "To learn more, check the README and FAQ"
	   );

    win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(win), PACKAGE " " VERSION);
    gtk_container_set_border_width(GTK_CONTAINER(win), 5);
    label = gtk_label_new(_(s));
    gtk_container_add(GTK_CONTAINER(win), label);
    gtk_widget_show_all(win);
}

/* SetWindowTitle() {{{1 */
void
SetWindowTitle(const char *title)
{
    char *p, *prev, *p1;
    char *buf;
    char tmpbuf[80];

    if (title)
    {
	strncpy(tmpbuf, title, sizeof(tmpbuf) - 1);
	tmpbuf[sizeof(tmpbuf) - 1] = '\0';
	buf = g_strconcat("", NULL);
	prev = tmpbuf;
	for (p = tmpbuf; (p = strchr(p, '%')) != NULL; p += 2)
	{
	    *p = '\0';
	    p1 = g_strconcat(buf, prev, NULL);
	    g_free(buf); buf = p1;
	    p1 = NULL;
	    switch (p[1])
	    {
		case 'p': p1 = g_strconcat(buf, PACKAGE, NULL);
			  break;
		case 'v': p1 = g_strconcat(buf, VERSION, NULL);
			  break;
		case 'n': if (ModemInfo)
			      p1 = g_strconcat(buf, ModemInfo->name, NULL);
			  else if (TelnetInfo)
			      p1 = g_strconcat(buf, TelnetInfo->name, NULL);
			  break;
		case 'h': if (ModemInfo)
			      p1 = g_strconcat(buf, ModemInfo->number, NULL);
			  else if (TelnetInfo)
			      p1 = g_strconcat(buf, TelnetInfo->host, NULL);
			  break;
		default: break;
	    }
	    if (p1 != NULL)
	    {
		g_free(buf);
		buf = p1;
	    }
	    prev = p + 2;
	}
	p1 = g_strconcat(buf, prev, NULL);
	g_free(buf); buf = p1;
    }
    else
	buf = g_strconcat(PACKAGE " - " VERSION, NULL);

    p1 = PC_IconvStr(buf, strlen(buf), 80);
    gtk_window_set_title(GTK_WINDOW(MainWin), p1);
    g_free(buf);
    if (p1 != buf)
	g_free(p1);
}

/* MenuOpenModem() {{{1 */
static void
MenuOpenModem(GtkWidget *w, GtkWidget *entry)
{
    const char *s;

    s = gtk_entry_get_text(GTK_ENTRY(entry));
    if (*s)
	ModemConnectByName(s);

    gtk_widget_destroy(MenuOpenWin);
    w = NULL;
}

/* MenuOpenTelnet() {{{1 */
static void
MenuOpenTelnet(GtkWidget *w, GtkWidget *entry)
{
    const char *s;

    s = gtk_entry_get_text(GTK_ENTRY(entry));
    if (*s)
	TelnetConnectByHostname(s);

    gtk_widget_destroy(MenuOpenWin);

    UNUSED(w);
}

/* MenuOpen() {{{1 */
static void
MenuOpen(void)
{
    GtkWidget *vbox, *hbox, *label, *entry, *button, *sep;

    MenuOpenWin = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_container_set_border_width(GTK_CONTAINER(MenuOpenWin), 5);
    gtk_window_set_title(GTK_WINDOW(MenuOpenWin), "Connect");
    gtk_window_set_position(GTK_WINDOW(MenuOpenWin), GTK_WIN_POS_MOUSE);

    vbox = gtk_vbox_new(FALSE, 2);
    gtk_container_add(GTK_CONTAINER(MenuOpenWin), vbox);

    hbox = gtk_hbox_new(FALSE, 5);
    gtk_container_add(GTK_CONTAINER(vbox), hbox);

    label = gtk_label_new("Host or Dial");
    gtk_container_add(GTK_CONTAINER(hbox), label);

    entry = gtk_entry_new();
    gtk_container_add(GTK_CONTAINER(hbox), entry);

    sep = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(vbox), sep, FALSE, FALSE, 5);

    hbox = gtk_hbox_new(FALSE, 10);
    gtk_container_add(GTK_CONTAINER(vbox), hbox);

    button = gtk_button_new_with_label("Modem");
    gtk_container_add(GTK_CONTAINER(hbox), button);
    gtk_signal_connect(GTK_OBJECT(button), "clicked",
		       (GtkSignalFunc) MenuOpenModem, entry);

    button = gtk_button_new_with_label("Telnet");
    gtk_container_add(GTK_CONTAINER(hbox), button);
    gtk_signal_connect(GTK_OBJECT(button), "clicked",
		       (GtkSignalFunc) MenuOpenTelnet, entry);

    button = gtk_button_new_with_label("Cancel");
    gtk_container_add(GTK_CONTAINER(hbox), button);
    gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
			      (GtkSignalFunc) gtk_widget_destroy,
			      (GtkObject *) MenuOpenWin);

    gtk_widget_show_all(MenuOpenWin);
    gtk_widget_grab_focus(entry);
}

/* MenuClose() {{{1 */
static void
MenuClose(void)
{
    if (PC_State == PC_RUN_STATE)
    {
	TRxStop();	/* file ̸  */
	PC_Hangup();
	PC_State = PC_IDLE_STATE;
    }
}

/* MenuQuit() {{{1 */
static void
MenuQuit(void)
{
    DoExit(0);
}

/* MenuResetTTY() {{{1 */
static void
MenuResetTTY(void)
{
    if (Term)
    {
	TermReset(Term);

	/* reset tty if modem is used */
	ModemResetTTY();
    }
}

/* MenuRunScript() {{{1 */
/* ScriptBox script file  μ ڰ
 *  ϴ script file Է  ֵ Ѵ.
 */
static void
MenuRunScript(void)
{
    if (ScriptRunning)
	ScriptCancel();
    else
    {
	const char *s = gtk_entry_get_text(GTK_ENTRY(ScriptBox));
	GeneralInputQuery(_("Script Run"), _("Filename:"), s,
			  (void (*)(char *)) ScriptRunFromFile, NULL);
    }
}

/* MenuRunEditor() {{{1 */
static void
MenuRunEditor(const char *file)
{
    char buf[MAX_PATH + 10];
    char *xeditor;

    if (file && *file)
    {
	xeditor = getenv("XEDITOR");
	if (xeditor == NULL || *xeditor == '\0')
	    xeditor = "gvim";
	g_snprintf(buf, sizeof(buf), "%s %s", xeditor, file);
	system(buf);
    }
}

/* MenuEditScript() {{{1 */
static void
MenuEditScript(void)
{
    const char *s;

    s = gtk_entry_get_text(GTK_ENTRY(ScriptBox));
    if (*s)
    {
	s = ScriptGetFullPath(s);
	if (s && *s)
	{
	    GeneralInputQuery(_("Script Edit"), _("Filename:"), s,
			      (void (*)(char *)) MenuRunEditor, NULL);
	}
    }
}

/* TB_Script() {{{1 */
static void
TB_Script(void)
{
    const char *s;

    if (ScriptRunning)
	ScriptCancel();
    else
    {
	s = gtk_entry_get_text(GTK_ENTRY(ScriptBox));
	if (*s)
	{
	    /* NOTE:  script  finish ʴ  -  
	     * waitfor  - combo box  key ޾ƸԴ´. ׷
	     * Term focus .
	     */
	    gtk_widget_grab_focus(GTK_WIDGET(Term));
	    ScriptRunFromFile(s);
	}
	else
	    GeneralInputQuery(_("Script Run"), _("Filename:"), NULL,
			      (void (*)(char *)) ScriptRunFromFile, NULL);
    }
}

/* MenuOptTblType {{{1 */
typedef struct _MenuOptTblType MenuOptTblType;
struct _MenuOptTblType {
    char *name;
    char *type;
};

/* MenuOptTbl {{{1 */
static MenuOptTblType MenuOptTbl[] = {
    {"Controlbar/View", "<ToggleItem>"},
    {"Toolbar/View", "<ToggleItem>"},
    {"Toolbar/Type/ICONS", ""},
    {"Toolbar/Type/TEXT", ""},
    {"Toolbar/Type/BOTH", ""},
    {"NewlineConvert/NONE", ""},
    {"NewlineConvert/CR", ""},
    {"NewlineConvert/LF", ""},
    {"NewlineConvert/CRLF", ""},
    {"Auto Response/Active", "<ToggleItem>"}
};

/* MenuOptionControl() {{{1 */
static void
MenuOptionControl(int inx)
{
    GtkWidget *w;
    char *path;

    path = g_strconcat("/Tools/Options/", MenuOptTbl[inx].name, NULL);
    w = gtk_item_factory_get_widget(MenuFactory, path);
    g_free(path);

    if (w == NULL)
    {
	g_warning("%s: cannot get widget for '%s'", __FUNCTION__,
		  MenuOptTbl[inx].name);
	return;
    }

    switch (inx)
    {
	case 0:	/* control-bar view */
	    /* control-bar hide false  control-bar  ̹Ƿ
	     * active ش.
	     */
	    gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(w),
					  ControlBarShow);
	    break;

	case 1:	/* tool-bar view */
	    /* tool-bar hide false  tool-bar  ̹Ƿ
	     * active ش.
	     */
	    gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(w), ToolBarShow);
	    break;

	case 9:	/* auto-response */
	    /* auto-response enable flag   ؼ toggleȴ. */
	    gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(w), TRUE);
	    break;

	default:
	    break;
    }
}

/* ControlBarFix() {{{1 */
void
ControlBarFix(void)
{
    if (CtrlBox)
    {
	if (ControlBarShow != GTK_WIDGET_VISIBLE(CtrlBox))
	{
	    if (ControlBarShow)
		gtk_widget_show(CtrlBox);
	    else
		gtk_widget_hide(CtrlBox);
	    MenuOptionControl(0);
	}
    }
}

/* WidgetShowHide() {{{1 */
static void
WidgetShowHide(GtkWidget *w, gboolean show)
{
    if (w)
    {
	if (show)
	{
	    if (!GTK_WIDGET_VISIBLE(w))
		gtk_widget_show(w);
	}
	else
	{
	    if (GTK_WIDGET_VISIBLE(w))
		gtk_widget_hide(w);
	}
    }
}

/* MenuOptions() {{{1 */
/* CtrlBox ToolBox  Ǳ   Լ ҷ  ִ , װ
 * MenuOptionControl() ؼ̴. menu active state ٲ  ֱ
 * ̴.
 */
static void
MenuOptions(const char *name)
{
    register int inx;

    for (inx = 0; inx < (int) ARYSIZ(MenuOptTbl); inx++)
	if (!strcmp(name, MenuOptTbl[inx].name))
	    break;

    switch (inx)
    {
	case 0:	/* control-bar view */
	    if (CtrlBox)
	    {
		ControlBarShow = !ControlBarShow;
		WidgetShowHide(CtrlBox, ControlBarShow);
	    }
	    break;

	case 1:	/* tool-bar view */
	    if (UseHandleBox)
	    {
		if (HandleBox)
		{
		    ToolBarShow = !ToolBarShow;
		    WidgetShowHide(HandleBox, ToolBarShow);
		}
	    }
	    else
	    {
		if (ToolBox)
		{
		    ToolBarShow = !ToolBarShow;
		    WidgetShowHide(ToolBox, ToolBarShow);
		}
	    }
	    break;

	case 2:
	    if (ToolBox)
	    {
		ScriptRunSizeChanged(FALSE);	/* small */
		gtk_toolbar_set_style(GTK_TOOLBAR(ToolBox), GTK_TOOLBAR_ICONS);
		g_free(ToolBarType);
		ToolBarType = g_strdup("ICONS");
	    }
	    break;

	case 3:
	    if (ToolBox)
	    {
		ScriptRunSizeChanged(FALSE);	/* small */
		gtk_toolbar_set_style(GTK_TOOLBAR(ToolBox), GTK_TOOLBAR_TEXT);
		g_free(ToolBarType);
		ToolBarType = g_strdup("TEXT");
	    }
	    break;

	case 4:
	    if (ToolBox)
	    {
		ScriptRunSizeChanged(TRUE);	/* large */
		gtk_toolbar_set_style(GTK_TOOLBAR(ToolBox), GTK_TOOLBAR_BOTH);
		g_free(ToolBarType);
		ToolBarType = g_strdup("BOTH");
	    }
	    break;

	    /* enter convert */
	case 5:
	    EnterConvert = ENTER_NONE;
	    break;
	case 6:
	    EnterConvert = ENTER_CR;
	    break;
	case 7:
	    EnterConvert = ENTER_LF;
	    break;
	case 8:
	    EnterConvert = ENTER_CRLF;
	    break;

	case 9:	/* auto response */
	    AutoResCheckEnabled = !AutoResCheckEnabled;
	    break;

	default:
	    g_warning("You may add some entry to MenuOptTbl.");
	    break;
    }
}

/* CreateNewPixmap() {{{1 */
static GtkWidget *
CreateNewPixmap(const char *filename)
{
    GtkWidget *w;
    GdkPixmap *pixmap;
    GdkBitmap *mask;
    char *f;

    w = NULL;

    /* check if I can access pixmap file */
    if ((f = GetRealFilename(filename, PKGDATADIR "/pixmap")) == NULL)
	return NULL;

    pixmap = gdk_pixmap_colormap_create_from_xpm(NULL,
						 gtk_widget_get_colormap
						 (MainWin), &mask,
						 &MainWin->style->
						 bg[GTK_STATE_NORMAL], f);
    g_free(f);

    if (pixmap)
    {
	w = gtk_pixmap_new(pixmap, mask);

	gdk_pixmap_unref(pixmap);
	gdk_pixmap_unref(mask);
    }

    return w;
}

/* CreateToolbar() {{{1 */
static GtkWidget *
CreateToolbar(GtkWidget *w, gint tbType)
{
    GtkWidget *tb, *pixmap, *cb;

#if GTK_MAJOR_VERSION >= 2
    tb = gtk_toolbar_new();
    gtk_toolbar_set_style(GTK_TOOLBAR(tb), tbType);
    gtk_toolbar_set_orientation(GTK_TOOLBAR(tb), GTK_ORIENTATION_HORIZONTAL);
#else
    tb = gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL, tbType);
    gtk_toolbar_set_button_relief(GTK_TOOLBAR(tb), GTK_RELIEF_NONE);
#endif

    if ((pixmap = CreateNewPixmap("tb_open.xpm")) != NULL)
	TB_OpenButton = gtk_toolbar_append_item(GTK_TOOLBAR(tb),
						_("Open"),
						_("Telnet or Modem connect"),
						NULL, pixmap,
						(GtkSignalFunc) MenuOpen, tb);

    if ((pixmap = CreateNewPixmap("tb_close.xpm")) != NULL)
	TB_CloseButton = gtk_toolbar_append_item(GTK_TOOLBAR(tb),
						 _("Close"),
						 _("Telnet or Modem disconnect"),
						 NULL, pixmap,
						 (GtkSignalFunc) MenuClose, tb);

    if ((pixmap = CreateNewPixmap("tb_exit.xpm")) != NULL)
	gtk_toolbar_append_item(GTK_TOOLBAR(tb),
				_("Exit"), _("Exit program"), NULL,
				pixmap, (GtkSignalFunc) MenuQuit, tb);

    if ((pixmap = ScriptRunAnimInit((GtkSignalFunc) TB_Script)) != NULL)
    {
	if ((cb = CreateScriptFileList(tb, TB_Script)) != NULL)
	{
	    gtk_toolbar_append_space(GTK_TOOLBAR(tb));

	    gtk_toolbar_append_widget(GTK_TOOLBAR(tb), cb, "Script file list",
				      NULL);

	    gtk_toolbar_append_space(GTK_TOOLBAR(tb));

	    gtk_toolbar_append_widget(GTK_TOOLBAR(tb), pixmap,
				      "Script file list", NULL);

	    gtk_toolbar_append_space(GTK_TOOLBAR(tb));
	}
	else
	    ScriptRunAnimDestroy();
    }

    if ((pixmap = CreateNewPixmap("tb_save.xpm")) != NULL)
	gtk_toolbar_append_item(GTK_TOOLBAR(tb),
				_("Capture"), _("Capture the input string"),
				NULL, pixmap, (GtkSignalFunc) CaptureControl,
				tb);

    if ((pixmap = CreateNewPixmap("tb_up.xpm")) != NULL)
	TB_UploadButton = gtk_toolbar_append_item(GTK_TOOLBAR(tb),
						  _("Upload"),
						  _("Upload a file"),
						  NULL, pixmap,
						  (GtkSignalFunc)
						  UploadFileQuery, tb);

    if ((pixmap = CreateNewPixmap("tb_down.xpm")) != NULL)
	TB_DownloadButton = gtk_toolbar_append_item(GTK_TOOLBAR(tb),
						    _("Download"),
						    _("Download a file"), NULL,
						    pixmap,
						    (GtkSignalFunc)
						    DownloadFileQuery, tb);

    return tb;

    w = NULL;	/* to make gcc happy */
}

/* PC_ButtonControl() {{{1 */
void
PC_ButtonControl(int state)
{
    gboolean open, close, up, down, trxStop;

    open = close = up = down = trxStop = FALSE;
    switch (state)
    {
	case TB_IDLE_STATE:
	    open = TRUE;
	    break;
	case TB_CONNECT_STATE:
	    close = up = down = TRUE;
	    break;
	case TB_TRX_STATE:
	    trxStop = TRUE;
	    break;
	default:
	    FatalExit(NULL, "invalid button sel.");
	    break;	/* never reach */
    }

    if (TB_OpenButton)
	gtk_widget_set_sensitive(TB_OpenButton, open);
    if (TB_CloseButton)
	gtk_widget_set_sensitive(TB_CloseButton, close);
    if (TB_UploadButton)
	gtk_widget_set_sensitive(TB_UploadButton, up);
    if (TB_DownloadButton)
	gtk_widget_set_sensitive(TB_DownloadButton, down);

    if (MenuOpenButton)
	gtk_widget_set_sensitive(MenuOpenButton, open);
    if (MenuCloseButton)
	gtk_widget_set_sensitive(MenuCloseButton, close);
    if (MenuUploadButton)
	gtk_widget_set_sensitive(MenuUploadButton, up);
    if (MenuDownloadButton)
	gtk_widget_set_sensitive(MenuDownloadButton, down);
    if (MenuStopTRxButton)
	gtk_widget_set_sensitive(MenuStopTRxButton, trxStop);
}

static char *
menuTranslate(const char *path, gpointer data)
{
    UNUSED(data);
    return _(path);
}

/* CreateMainWindow() {{{1 */
static void
CreateMainWindow(void)
{
    register int i, tbType;
    GtkWidget *hbox, *frame, *sw, *sep;
    GtkAccelGroup *accelGroup;
    GdkGeometry geom;
    GdkWindowHints geomMask;
    static const GtkTargetEntry targetEntry[] = {
	{"STRING", 0, 0}
    };
    GtkItemFactoryEntry entry;
    /* NOTE: Terminal emulation α׷̹Ƿ ޴ ۼ 
     *  ʿ䰡 ִ. Ư Control Ű    ̸
     * Ϲ Alt key ϴٰ ϰڴ.
     *
     * Used hotkey list:
     * Alt: c d e f h l o p r s t u w x
     */
    static GtkItemFactoryEntry menuItems[] = {
	{N_("/_File"), NULL, 0, 0, _GN("<Branch>")},
	{N_("/_File/_Open"), "<alt>o", MenuOpen, 0, _GS(GTK_STOCK_OPEN)},
	{N_("/_File/_Close"), NULL, MenuClose, 0, _GS(GTK_STOCK_CLOSE)},
	{("/_File/---"), NULL, 0, 0, _GN("<Separator>")},
	{N_("/_File/_Exit"), "<alt>x", MenuQuit, 0, _GS(GTK_STOCK_QUIT)},
	{N_("/_Tools"), NULL, 0, 0, _GN("<Branch>")},
	{N_("/_Tools/_Options"), NULL, 0, 0, _GN("<Branch>")},
	{N_("/_Tools/_Options/Controlbar"), NULL, 0, 0, _GN("<Branch>")},
	{N_("/_Tools/_Options/Toolbar"), NULL, 0, 0, _GN("<Branch>")},
	{N_("/_Tools/_Options/Toolbar/Type"), NULL, 0, 0, _GN("<Branch>")},
	{N_("/_Tools/_Options/NewlineConvert"), NULL, 0, 0, _GN("<Branch>")},
	{N_("/_Tools/_Options/Auto Response"), NULL, 0, 0, _GN("<Branch>")},
	{("/_Tools/---"), NULL, 0, 0, _GN("<Separator>")},
	{N_("/_Tools/_Download"), "<alt>d",
	 (GtkItemFactoryCallback) DownloadFileQuery, 0, _GN(NULL)},
	{N_("/_Tools/_Upload"), "<alt>p",
	 (GtkItemFactoryCallback) UploadFileQuery, 0, _GN(NULL)},
	{N_("/_Tools/_Stop Transfer"), "<alt>r", TRxStop, 0, _GN(NULL)},
	{("/_Tools/---"), NULL, 0, 0, _GN("<Separator>")},
	{N_("/_Tools/_Capture"), "<alt>l",
	 (GtkItemFactoryCallback) CaptureControl, 0, _GN(NULL)},
	{N_("/_Tools/Script _Run-Stop"), "<alt>s", MenuRunScript, 0, _GN(NULL)},
	{N_("/_Tools/Script _Edit"), "<alt>e", MenuEditScript, 0, _GN(NULL)},
	{N_("/_Tools/Cha_t"), "<alt>c", MenuChat, 0, _GN(NULL)},
	{N_("/_Tools/Reset Ter_minal"), "<alt>u", MenuResetTTY, 0,
	    _GS(GTK_STOCK_CLEAR)},
	{N_("/_Tools/Open _History Window"), "<alt>w", ModemOpenHistoryWin,
	    0, _GN(NULL)},
	{N_("/_Bookmarks"), NULL, 0, 0, _GN("<Branch>")}
    };
    static GtkItemFactoryEntry helpEntry[] = {
	{N_("/_Help"), NULL, 0, 0, _GN("<LastBranch>")},
	{N_("/_Help/_About..."), NULL, aboutDialog, 0, _GS(GTK_STOCK_HELP)}
    };

    MainWin = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_policy(GTK_WINDOW(MainWin), TRUE, TRUE, TRUE);
    gtk_signal_connect(GTK_OBJECT(MainWin), "destroy",
		       GTK_SIGNAL_FUNC(MenuQuit), NULL);
    SetWindowTitle(NULL);
    gtk_widget_set_name(MainWin, "main");	/* for gtkrc */
    gtk_container_set_border_width(GTK_CONTAINER(MainWin), 5);

    MainBox = gtk_vbox_new(FALSE, 3);
    gtk_container_add(GTK_CONTAINER(MainWin), MainBox);
    gtk_widget_show(MainBox);

    accelGroup = gtk_accel_group_new();
    MenuFactory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", accelGroup);

    gtk_item_factory_set_translate_func(MenuFactory, menuTranslate, NULL, NULL);
    gtk_item_factory_create_items(MenuFactory, ARYSIZ(menuItems), menuItems,
				  NULL);

    AddBookmarkMenus(MenuFactory, (void (*)(gpointer)) ModemConnectByName,
		     (void (*)(gpointer)) TelnetConnectByHostname);

    gtk_item_factory_create_items(MenuFactory, ARYSIZ(helpEntry), helpEntry,
				  NULL);

    memset(&entry, 0, sizeof(entry));
    for (i = 0; i < (int) ARYSIZ(MenuOptTbl); i++)
    {
	entry.path = g_strconcat(N_("/_Tools/_Options/"), MenuOptTbl[i].name, NULL);
	entry.item_type = MenuOptTbl[i].type;
	entry.callback = MenuOptions;
	gtk_item_factory_create_item(MenuFactory, &entry,
				     (gpointer) MenuOptTbl[i].name, 1);

	MenuOptionControl(i);
	g_free(entry.path);
    }

    gtk_window_add_accel_group(GTK_WINDOW(MainWin), accelGroup);
    gtk_box_pack_start(GTK_BOX(MainBox), MenuFactory->widget, FALSE, FALSE, 0);
    gtk_widget_show(MenuFactory->widget);

    /* menu active ¸ ϱ   ´. */
    MenuOpenButton = gtk_item_factory_get_widget(MenuFactory,
						 N_("/_File/_Open"));
    MenuCloseButton = gtk_item_factory_get_widget(MenuFactory,
						  N_("/_File/_Close"));
    MenuDownloadButton = gtk_item_factory_get_widget(MenuFactory,
						     N_("/_Tools/_Download"));
    MenuStopTRxButton = gtk_item_factory_get_widget(MenuFactory,
						    N_("/_Tools/_Stop Transfer"));
    MenuUploadButton = gtk_item_factory_get_widget(MenuFactory,
						   N_("/_Tools/_Upload"));

    /* toolbar */

    if (!strcasecmp(ToolBarType, "ICONS"))
	tbType = GTK_TOOLBAR_ICONS;
    else if (!strcasecmp(ToolBarType, "TEXT"))
	tbType = GTK_TOOLBAR_TEXT;
    else
    {
	if (strcasecmp(ToolBarType, "BOTH"))
	{
	    g_free(ToolBarType);
	    ToolBarType = g_strdup("BOTH");
	}
	tbType = GTK_TOOLBAR_BOTH;
    }

    if (UseHandleBox)
    {
	HandleBox = gtk_handle_box_new();
	gtk_box_pack_start(GTK_BOX(MainBox), HandleBox, FALSE, FALSE, 0);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(HandleBox), hbox);
	gtk_container_set_border_width(GTK_CONTAINER(HandleBox), 3);

	if ((ToolBox = CreateToolbar(hbox, tbType)) != NULL)
	{
	    gtk_box_pack_start(GTK_BOX(hbox), ToolBox, FALSE, FALSE, 0);
	    gtk_widget_show_all(HandleBox);
	    if (!ToolBarShow)
		gtk_widget_hide(HandleBox);
	}
    }
    else
    {
	if ((ToolBox = CreateToolbar(MainBox, tbType)) != NULL)
	{
	    gtk_box_pack_start(GTK_BOX(MainBox), ToolBox, FALSE, FALSE, 0);
	    gtk_widget_show_all(ToolBox);
	    if (!ToolBarShow)
		gtk_widget_hide(ToolBox);
	}
    }

    /* terminal */

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(MainBox), hbox);

    frame = gtk_frame_new(NULL);
    gtk_container_set_border_width(GTK_CONTAINER(frame), 0);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_OUT);
    gtk_container_add(GTK_CONTAINER(hbox), frame);

    TermBox = gtk_vbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(frame), TermBox);

    sw = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), GTK_POLICY_NEVER,
				   GTK_POLICY_ALWAYS);
    gtk_container_add(GTK_CONTAINER(TermBox), sw);

    Term = TermNew();
#if GTK_MAJOR_VERSION < 2
    gtk_accel_group_attach(accelGroup, GTK_OBJECT(Term));
#else
    _gtk_accel_group_attach(accelGroup, G_OBJECT(Term));
#endif
    gtk_widget_set_usize(GTK_WIDGET(Term), TermCol * Term->fontWidth,
			 TermRow * Term->fontHeight);
    gtk_container_add(GTK_CONTAINER(sw), GTK_WIDGET(Term));
    gtk_widget_show_all(hbox);

    /* geometry */
    geomMask = GDK_HINT_MIN_SIZE | GDK_HINT_BASE_SIZE | GDK_HINT_RESIZE_INC;
    geom.min_width = Term->fontWidth * TermCol;
    geom.min_height = Term->fontHeight * TermRow;
    geom.base_width = 0;
    geom.base_height = 0;
    geom.width_inc = Term->fontWidth;
    geom.height_inc = Term->fontHeight;
    gtk_window_set_geometry_hints(GTK_WINDOW(MainWin), GTK_WIDGET(Term),
				  &geom, geomMask);

    gtk_selection_add_targets(GTK_WIDGET(Term), GDK_SELECTION_PRIMARY,
			      targetEntry,
			      sizeof(targetEntry) / sizeof(targetEntry[0]));

    gtk_widget_grab_focus(GTK_WIDGET(Term));

    CtrlBox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(MainBox), CtrlBox, FALSE, FALSE, 0);
    gtk_container_set_border_width(GTK_CONTAINER(CtrlBox), 0);
    gtk_widget_show_all(CtrlBox);
    if (!ControlBarShow)
	gtk_widget_hide(CtrlBox);

    sep = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(MainBox), sep, FALSE, FALSE, 0);
    gtk_widget_show_all(sep);

    StatBox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(MainBox), StatBox, FALSE, FALSE, 0);
    CreateStatusBar(StatBox);
    gtk_widget_show_all(StatBox);

    PC_ButtonControl(TB_IDLE_STATE);

    gtk_widget_show(MainWin);

    /* use xterm cursor */
    {
	GdkCursor *cursor = gdk_cursor_new(GDK_XTERM);
	gdk_window_set_cursor(Term->widget.window, cursor);
	gdk_cursor_destroy(cursor);
    }
}

/* FatalExit() {{{1 */
void
FatalExit(GtkWidget *parent, const gchar *form, ...)
{
    gchar *p, *p2;
    va_list args;
    GtkWidget *w;

    va_start(args, form);
    p = g_strdup_vprintf(form, args);
    va_end(args);

    p2 = g_strconcat("\n   ", p, "   \n", NULL);
    w = DialogShowMessage(parent, PACKAGE ": FATAL ERROR", "EXIT", p2);
    g_free(p);
    g_free(p2);
    if (w)
    {
	gtk_window_set_modal(GTK_WINDOW(w), TRUE);
	gtk_signal_connect(GTK_OBJECT(w), "destroy",
			   GTK_SIGNAL_FUNC(MenuQuit), NULL);

	/* ̻  ۾  ! */
	gtk_main();
    }
    DoExit(0);
}

/* main() {{{1 */
int
main(int argc, char *argv[])
{
    register int i;
    char *conName = NULL;

#ifdef HAVE_GETTEXT
    bindtextdomain (PACKAGE, LOCALE_DIR);
#if GTK_MAJOR_VERSION >= 2
    bind_textdomain_codeset(PACKAGE, "UTF-8");
#endif
    textdomain (PACKAGE);
    setlocale (LC_ALL, "");
#endif

    gtk_set_locale();
    gtk_init(&argc, &argv);
#if GTK_MAJOR_VERSION < 2
    gdk_imlib_init();
#endif
    g_thread_init(NULL);

    MyFullName = argv[0];
    MyName = g_basename(MyFullName);
    signal(SIGSEGV, caughtSignal);
    signal(SIGKILL, caughtSignal);

    ResourceCheckInstall();

    RegisterChildSignalHandler();

#if defined(_DEBUG_)
    for (i = 0; i < MAX_OPRT_NUM; i++)
	OPrintFlag[i] = FALSE;
#endif

    DoConfig();

    if (IconvFromStr != NULL && IconvToStr != NULL)
	PC_IconvOpen(IconvToStr, IconvFromStr);

    for (i = 1; i < argc; i++)
    {
	if (argv[i][0] == '-')
	{
	    if (!strcmp(argv[i], "--version"))
	    {
		printf("%s\n", VERSION);
		exit(0);
	    }
	    else if (!strcmp(argv[i], "--modem"))
		UseModem = TRUE;
	    else if (!strcmp(argv[i], "--telnet"))
		UseModem = FALSE;
#ifdef _DEBUG_
	    else if (!strcmp(argv[i], "--debug") && (i + 1) < argc)
	    {
		unsigned opt;
		char *s, *s2;

		++i;
		for (s = argv[i]; *s; s = s2 + 1)
		{
		    if ((s2 = strchr(s, ',')) != NULL)
			*s2 = '\0';

		    if (sscanf(s, "%u", &opt) == 1 && opt < ARYSIZ(OPrintFlag))
			OPrintFlag[opt] = TRUE;

		    if (s2 == NULL)
			break;
		}
	    }
#endif
	    else if (!strcmp(argv[i], "--help"))
	    {
		printf("Usage: %s <options> <hostname|dial_num>\n"
#ifdef _DEBUG_
		       "  --debug #n,#n1,... : enable debug\n"
#endif
		       "  --help	         : this help message\n"
		       "  --modem            : use modem if dial_num given\n"
		       "  --telnet           : use telnet if hostname given\n"
		       "  --version          : display version\n", MyName);
		exit(0);
	    }
	    else
	    {
		printf("To know valid option, type '%s --help'.\n", MyName);
		exit(0);
	    }
	}
	else
	    conName = argv[i];
    }

    CreateMainWindow();

    TermEmulModeSet();
    AutoResponseInit();
    I_SelectInit();
    TRxInit();

    ScriptRunFromFile(ExpandPath(DEFAULT_STARTUP_SCRIPT));

    SoundPlay(SoundStartFile);

    if (conName)
    {
	if (UseModem)
	    gtk_timeout_add(500, (GtkFunction) ModemConnectByName, conName);
	else
	    gtk_timeout_add(500, (GtkFunction) TelnetConnectByHostname, conName);
    }

    gtk_main();

    DoExitClean();

    return 0;
}
