/*
 * Script
 *
 * Copyright (C) 2000-2002, Nam SungHyun and various contributors
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#include "config.h"
#include <ctype.h>	/* tolower(), isdigit() */
#include <sys/stat.h>
#include <glob.h>	/* glob() */
#include <fcntl.h>	/* X_OK */
#if GTK_MAJOR_VERSION < 2
#include <gdk_imlib.h>	/* gdk_imlib_get_visual() */
#else
#include <gdk-pixbuf/gdk-pixbuf.h>
#endif

#include "pcMain.h"
#include "pcTerm.h"

/* definitions & declarations {{{1 */

/* ȥ㸻 ũƮ ΰ? */
gboolean ScriptRunning = FALSE;

char *ScriptPath;
char *ScriptAnimFile;

/* script enum {{{1 */
enum {
    SCR_NONE,		/* nothing */
    SCR_ASSIGN, 	/* assign Ưϰ ó */
    SCR_AUTORES,	/* auto response file  */
    SCR_BEEP,		/*  Ҹ  */
    SCR_BOOKMARK,	/* bookmark menu ߰ */
    SCR_BUTTON,		/* control-bar button ߰ */
    SCR_CAPTURE,	/* capture on or off */
    SCR_CLEAR,		/* ȭ clear */
    SCR_CTLMENU,	/* control-bar menu ߰ */
    SCR_DISCONNECT,	/*   */
    SCR_DOWNLOAD,	/* file download */
    SCR_ELSE,		/* else */
    SCR_ELSEIF,		/* elseif/elif */
    SCR_EMULATE,	/* terminal emulation mode change */
    SCR_END,		/* script end */
    SCR_ENDIF,		/* if ǹ end */
    SCR_ENDWIN,		/* script end window */
    SCR_GOTO,		/* goto label */
    SCR_GOTO_TAG,	/* goto tag */
    SCR_IF,		/* if ǹ */
    SCR_INCLUDE,	/* include another script file */
    SCR_KFLUSH,		/* key buffer flush */
    SCR_LABEL,		/* control-bar label ߰ */
    SCR_MESSAGE,	/* ȭ鿡 ޽  */
    SCR_RUN,		/* external command execute */
    SCR_PAUSE,		/* pause seconds */
    SCR_RFLUSH,		/* rx buffer flush */
    SCR_TRANSMIT,	/* transmit string */
    SCR_UPLOAD,		/* file upload */
    SCR_WAITFOR,	/* wait string */

    /* set */
    SCR_SET,		/* general parameter setting */
    SCR_SET_BAUDRATE,	/* modem baudrate change */
    SCR_SET_CONFIG,	/* config variable change */
    SCR_SET_DEBUG,	/* internal debugging option */
    SCR_SET_PROTOCOL,	/* transfer protocol change */
    SCR_SET_TITLE,	/* â   */

    /* get */
    SCR_GET,
    SCR_GET_BAUDRATE,	/* baudrate get */
    SCR_GET_FILEINFO,	/* file info get */

    SCR_WIN,		/* â */
    SCR_WIN_WINDOW,	/* â  */
    SCR_WIN_HBOX,	/*   */
    SCR_WIN_VBOX,	/*   */
    SCR_WIN_FRAME,	/*  */
    SCR_WIN_LABEL,	/*  */
    SCR_WIN_ENTRY,	/* Է â */
    SCR_WIN_BUTTON,	/*  */
    SCR_WIN_HSEP,	/*  */
    SCR_WIN_END,	/* â   (script ) */
    SCR_WIN_WAIT,	/* â   (â   script ) */

    MAX_SCRIPT_TYPE
};

/* ScriptTableType {{{1 */
typedef struct _ScriptTableType ScriptTableType;

struct _ScriptTableType {
    int type;
    char *token;
    int argc;
    const ScriptTableType *next;
};

/* table argc麸 1 ̻ ĿѴ. */
#define MAX_SCRIPT_PARAMS  6

/* NOTE: argc field token ߸  ּ Ķ ̴.
 * ̰ script parser Ͽ ش token  ڰ script 
 * ۼϿ  ˻簡 ϵ ش.
 *  , 'win begin' parameter  ǰ, <title> parameter
 * ־ Ǵ 찡 ̿ شǴ , ̶ argc field 0 ־ Ѵ.
 * script parser MAX_SCRIPT_PARAMS  ScriptType argc  argv[]
 *  ֹǷ, ߿ <title> ִ  ˻  ִ.
 */

/* ScriptSetTable {{{1 */
static const ScriptTableType ScriptSetTable[] = {
    {SCR_SET_BAUDRATE, "baudrate", 1, NULL},	/* set baudrate 19200 */
    {SCR_SET_CONFIG, "config", 2, NULL},	/* set config var value */
    {SCR_SET_DEBUG, "debug", 1, NULL},	/* set debug option# */
    {SCR_SET_PROTOCOL, "protocol", 1, NULL},	/* set protocol [ascii/raw-ascii/kermit/xmodem/ymodem/zmodem] */
    {SCR_SET_TITLE, "title", 1, NULL},	/* set title "string" */
    {-1, NULL, -1, NULL}	/* end of table */
};

/* ScriptGetTable {{{1 */
static const ScriptTableType ScriptGetTable[] = {
    {SCR_GET_BAUDRATE, "baudrate", 1, NULL},	/* get baudrate $var */
    {SCR_GET_FILEINFO, "file", 3, NULL }, /* get file filename size $var */
    {-1, NULL, -1, NULL}	/* end of table */
};

/* ScriptWinTable {{{1 */
static const ScriptTableType ScriptWinTable[] = {
    {SCR_WIN_WINDOW, "begin", 0, NULL},	/* win begin <title> */
    {SCR_WIN_HBOX, "hbox", 3, NULL},	/* win hbox w name 3 */
    {SCR_WIN_VBOX, "vbox", 3, NULL},	/* win vbox w name 5 */
    {SCR_WIN_FRAME, "frame", 3, NULL }, /* win frame w name title */
    {SCR_WIN_LABEL, "label", 2, NULL }, /* win label w name */
    {SCR_WIN_ENTRY, "entry", 2, NULL},	/* win entry w $STRTAG <act button-title> */
    {SCR_WIN_BUTTON, "button", 4, NULL},	/* win button w title [string/script/sfile] "string-script-file" */
    {SCR_WIN_HSEP, "hsep", 1, NULL},	/* win hsep w */
    {SCR_WIN_END, "end", 0, NULL},	/* win end */
    {SCR_WIN_WAIT, "wait", 0, NULL},	/* win wait */
    {-1, NULL, -1, NULL}	/* end of table */
};

/* ScriptMainTable {{{1 */
static const ScriptTableType ScriptMainTable[] = {
    /* assign S? "string"   (? = number).
     * assign label  Ưϰ óѴ.
     */
    {SCR_ASSIGN, "assign", 2, NULL },
    {SCR_AUTORES, "autores", 4, NULL},	/* autores add "find" [string|script] "string" */
    {SCR_BEEP, "beep", 0, NULL},	/* beep <intensity> */
    {SCR_BOOKMARK, "bookmark", 4, NULL},	/* bookmark [add/gadd] "name" [string|script] "string" */
    {SCR_BUTTON, "button", 4, NULL},	/* button [add/gadd] "name" [string|script] "string" */
    {SCR_CAPTURE, "capture", 1, NULL},	/* capture [on/off] <filename> <includebuf> */
    {SCR_CLEAR, "clear", 0, NULL},	/* clear */
    {SCR_CTLMENU, "ctlmenu", 5, NULL},	/* ctlmenu [add/gadd] title name [string/script] "string-action" */
    {SCR_DISCONNECT, "disconnect", 0, NULL},	/* disconnect */
    {SCR_DOWNLOAD, "receivefile", 1, NULL},	/* receivefile xmodem filename */
    {SCR_ELSE, "else", 0, NULL},	/* else */
    {SCR_ELSEIF, "elseif", 3, NULL},	/* elseif $var == "val" */
    {SCR_ELSEIF, "elif", 3, NULL},	/* elif $var == "val" */
    {SCR_EMULATE, "emulate", 1, NULL},	/* emulate vt100/ansi/.. */
    {SCR_END, "end", 0, NULL},	/* end */
    {SCR_ENDIF, "fi", 0, NULL},	/* fi */
    {SCR_ENDIF, "endif", 0, NULL},	/* endif */
    {SCR_ENDWIN, "endwin", 0, NULL},	/* end window */
    {SCR_GOTO, "goto", 1, NULL},	/* goto label */
    /* SCR_GOTO_TAG Ư token . xxxx:  :  tag  */
    {SCR_IF, "if", 3, NULL},	/* if $? == 0 */
    {SCR_INCLUDE, "include", 1, NULL},	/* include "script-file" */
    {SCR_KFLUSH, "kflush", 0, NULL},	/* kflush */
    {SCR_LABEL, "label", 2, NULL}, /* label [add/gadd/rmv] [baudrate/protocol/emulate/linestatus] */
    {SCR_MESSAGE, "message", 1, NULL},	/* message "string" */
    {SCR_RUN, "run", 1, NULL},	/* run 'filename' */
    {SCR_SET, "set", 1, ScriptSetTable},	/* set ??? */
    {SCR_GET, "get", 1, ScriptGetTable}, 	/* get ??? */
    {SCR_PAUSE, "pause", 1, NULL},	/* pause secs */
    {SCR_RFLUSH, "rflush", 0, NULL},	/* rflush */
    {SCR_TRANSMIT, "transmit", 1, NULL},	/* transmit "string" */
    {SCR_UPLOAD, "sendfile", 2, NULL},	/* sendfile zmodem filename */
    {SCR_WAITFOR, "waitfor", 1, NULL},	/* waitfor "string" <timeout> <"str2"> */
    {SCR_WIN, "win", 1, ScriptWinTable}, /* win ??? */
    {-1, NULL, -1, NULL}	/* end of table */
};

/* ScriptType {{{1 */
/* ScriptBookmarkType/ButtonRemoveType type */
#define SB_SCRIPT	   0x01	/* script ڿ?  bit 0̸ string */
#define SB_SCRIPTFILE	   0x02	/* script file? */
#define SB_BRANCH	   0x02	/* bookmark  branch ƴ.. */
#define SB_GLOBAL	   0x04	/* disconnect remove or not */

typedef struct {
    int type;
    int argc;
    gpointer *argv;
    gpointer *savedArgv;
} ScriptType;

/* AssignType {{{1 */
typedef struct {
    char *var;
    char *val;
} AssignType;

/* Local variables {{{1 */

/* script list ScriptBox ϱ  gtk_combo_set_popdown_strings()
 * Լ ̿ϴ , GList̾ Ѵ.
 */
static GList *CurrScriptItem = NULL;	/*   script item */
static GList *ScriptItemList = NULL;	/* ü script item list */

/* Script   ߴܵ  CurrScriptItem  ´. */
static GList *ScriptWinSavedList = NULL;

static char *CurrScriptFile = NULL;
static GSList *AssignList = NULL;
static gint ScriptResult = OK;
static gboolean WaitForPending = FALSE;
static guint WaitForTimeoutTag = 0;
static char *WaitForString = NULL;
static char *WaitForString2 = NULL;
static guint IdleTag = 0;
static guint PauseTag = 0;
static gboolean ScriptCaptureActive = FALSE;

static char *ScriptChildStr = NULL;

static GSList *ScriptLabelList = NULL;
static gboolean ScriptLabelRemoved = TRUE;

/* ScriptWinObjType {{{1 */
typedef struct {
    char *str;
    char *env;
    int type;
} ScriptWinObjType;

static GSList *ScriptWinObjList = NULL;

/* 'win' ɿ   â */
static GtkWidget *ScriptWin = NULL;

/* ScriptBookmarkType {{{1 */
typedef struct {
    char *id;
    int type;
    char *str;
} ScriptBookmarkType;

static GSList *BookmarkList = NULL;
static gboolean BookmarkRemoved = TRUE;

/* ScriptButtonType {{{1 */
typedef struct {
    char *id;
    int type;
    char *str;
} ScriptButtonType;

static GSList *ButtonList = NULL;
static gboolean ButtonRemoved = TRUE;

/* script execute command */
static int ScriptRunFD[2];
static guint ScriptRunInputTag;

/* Local function prototypes {{{1 */

static GList *ScriptMakeListFromFile(char *filename);
static int ScriptRunNext(void);
static void ScriptRunAnimStart(void);
static void ScriptRunAnimStop(void);
static AssignType *GetAssign(const char *s);
static int WaitForInputFilter(char *buf, int len);
static int AddAssign(char *var, char *val);

static void ScriptRemoveLabel(void);
static void ScriptRemoveBookmark(void);
static void ScriptRemoveButton(void);
static void ScriptDestroyWinObjList(void);

/* ScriptDestroy() {{{1 */
static void
ScriptDestroy(void)
{
    register int i;

    CurrScriptItem = NULL;
    ScriptWinSavedList = NULL;

    if (ScriptItemList)
    {
	GList *list;
	ScriptType *script;

	/* destroy script list */
	for (list = ScriptItemList; list; list = list->next)
	{
	    script = list->data;
	    if (script->argc > 0)
	    {
		for (i = 0; i < script->argc; i++)
		    g_free(script->argv[i]);
		g_free(script->argv);
		if (script->savedArgv)
		{
		    for (i = 0; i < script->argc; i++)
			if (script->savedArgv[i])
			    g_free(script->savedArgv[i]);
		    g_free(script->savedArgv);
		}
	    }
	    g_free(script);
	}
	g_list_free(ScriptItemList);
	ScriptItemList = NULL;
    }

    if (CurrScriptFile)
    {
	g_free(CurrScriptFile);
	CurrScriptFile = NULL;
    }

    /* destroy assign list */
    if (AssignList)
    {
	GSList *slist;
	AssignType *assign;

	for (slist = AssignList; slist; slist = slist->next)
	{
	    assign = slist->data;
	    g_free(assign->var);
	    g_free(assign->val);
	}
	g_slist_free(AssignList);
	AssignList = NULL;
    }
}

/* ScriptRemoveAll() {{{1 */
static void
ScriptRemoveAll(void)
{
    if (WaitForPending)
    {
	WaitForPending = FALSE;
	InputFilterList = g_slist_remove(InputFilterList, WaitForInputFilter);
    }
    if (WaitForTimeoutTag > 0)
    {
	gtk_timeout_remove(WaitForTimeoutTag);
	WaitForTimeoutTag = 0;
    }
    if (IdleTag > 0)
    {
	gtk_timeout_remove(IdleTag);
	IdleTag = 0;
    }
    if (PauseTag > 0)
    {
	gtk_timeout_remove(PauseTag);
	PauseTag = 0;
    }
    if (ScriptCaptureActive)
    {
	ScriptCaptureActive = FALSE;
	CaptureFinish();
    }
    ScriptDestroy();
    ScriptRunAnimStop();
    ScriptRunning = FALSE;
}

/* ScriptFinish() {{{1 */
static gint
ScriptFinish(void)
{
    if (CurrScriptFile)
	StatusShowMessage(_("Script '%s' finished."), CurrScriptFile);
    ScriptRemoveAll();
    ScriptResult = OK;
    return OK;
}

/* ScriptExit() {{{1 */
void
ScriptExit(void)
{
    ScriptFinish();
    ScriptRemoveLabel();
    ScriptRemoveBookmark();
    ScriptRemoveButton();
    ScriptDestroyWinObjList();
}

/* ScriptInstallNextItem() {{{1 */
static void
ScriptInstallNextItem(void)
{
    if (TRxRunning)
	PC_Sleep(1);
    if (CurrScriptItem == ScriptWinSavedList)
	/* will install after window destroyed */
	return;
    IdleTag = gtk_timeout_add(0, (GtkFunction) ScriptRunNext, NULL);
}

/* WaitForInputFilter() {{{1 */
static int
WaitForInputFilter(char *buf, int len)
{
    register int i;
    static char *p, *p2;

    if (len < 0)
    {
	p = WaitForString;
	p2 = WaitForString2;
	return 0;
    }

    for (i = 0; i < len; i++)
    {
	if (buf[i] != *p)
	{
	    p = WaitForString;
	    if (WaitForString2 && buf[i] != *p2)
		p2 = WaitForString2;
	    else
		p2++;

	}
	else
	    p++;
	
	if (*p == '\0' || (WaitForString2 && *WaitForString2 && *p2 == '\0'))
	{
	    /* match string */
	    StatusShowMessage(" ");	/* clear 'Waitfor...' if exists */
	    InputFilterList =
		g_slist_remove(InputFilterList, WaitForInputFilter);
	    WaitForPending = FALSE;
	    if (WaitForTimeoutTag > 0)
	    {
		gtk_timeout_remove(WaitForTimeoutTag);
		WaitForTimeoutTag = 0;
	    }

	    if (*p == '\0')
		ScriptResult = 0;
	    else
		ScriptResult = 1;
	    if (CurrScriptItem && CurrScriptItem->next)
	    {
		CurrScriptItem = CurrScriptItem->next;
		ScriptInstallNextItem();
	    }
	    else
		ScriptFinish();
	}
    }
    return len;
}

/* WaitForTimer() {{{1 */
static int
WaitForTimer(gpointer data)
{
    UNUSED(data);

    WaitForPending = FALSE;
    InputFilterList = g_slist_remove(InputFilterList, WaitForInputFilter);

    ScriptResult = ERROR;
    WaitForTimeoutTag = 0;
    if (CurrScriptItem && CurrScriptItem->next)
    {
	CurrScriptItem = CurrScriptItem->next;
	ScriptInstallNextItem();
    }
    else
	ScriptFinish();
    StatusShowMessage(_("waitfor timeout!"));
    return FALSE;
}

/* ScriptWaitFor() {{{1 */
static void
ScriptWaitFor(ScriptType *script, float seconds)
{
    if (WaitForString)
	g_free(WaitForString);
    WaitForString = ConvertCtrlChar((char *) script->argv[0]);

    if (WaitForString2)
    {
	g_free(WaitForString2);
	WaitForString2 = NULL;
    }
    if (script->argc >= 3)
	WaitForString2 = ConvertCtrlChar((char *) script->argv[2]);

    WaitForInputFilter(NULL, -1);	/* initialize the static string ptr */

    WaitForPending = TRUE;

    InputFilterList = g_slist_append(InputFilterList, WaitForInputFilter);

    StatusShowMessage(_("Waiting '%s'..."), WaitForString);

    if (seconds >= 0.1)
	WaitForTimeoutTag = gtk_timeout_add((guint32) (seconds * 1000UL),
					    WaitForTimer, NULL);
}

/* ScriptGetBaudrate() {{{1 */
static gint
ScriptGetBaudrate(ScriptType *s)
{
    guint32 b = ModemGetSpeed();
    char buf[20];

    g_snprintf(buf, sizeof(buf), "%lu", (unsigned long) b);
    return AddAssign(s->argv[0], buf);
}

/* ScriptGetFileInfo() {{{1 */
static gint
ScriptGetFileInfo(ScriptType *s)
{
    gint sts = ERROR;
    struct stat t;
    char *fname;
    char *info = (char *) s->argv[1];
    char *vars = (char *) s->argv[2];
    char buf[32];

    if ((fname = GetRealFilename((char *) s->argv[0], UploadPath)) == NULL)
    {
	g_warning("Cannot access '%s'.", (char *) s->argv[0]);
	return ERROR;
    }

    if (stat(fname, &t))
	g_warning("Cannot get stat for %s\n", fname);
    else
    {
	if (!strcasecmp(info, "size"))
	{
	    snprintf(buf, sizeof(buf), "%lu", (unsigned long) t.st_size);
	    sts = AddAssign(vars, buf);
	}
	else
	    g_warning("Unknown info name %s\n", info);
    }

    g_free(fname);
    return sts;
}

/* ScriptGoto() {{{1 */
static gint
ScriptGoto(char *label)
{
    GList *list;
    ScriptType *script;

    for (list = ScriptItemList; list; list = list->next)
    {
	script = list->data;
	if (script->type == SCR_GOTO_TAG)
	{
	    if (!strcasecmp(script->argv[0], label))
	    {
		CurrScriptItem = list;
		return OK;
	    }
	}
    }

    return ERROR;
}

/* ScriptRunBookmark() {{{1 */
static void
ScriptRunBookmark(ScriptBookmarkType * sb)
{
    if (sb->type & SB_SCRIPT)
	ScriptRunFromString(sb->str);
    else if (sb->type & SB_SCRIPTFILE)
	ScriptRunFromFile(sb->str);
    else
	Term->remoteWrite(sb->str, strlen(sb->str));
}

/* ScriptRunButton() {{{1 */
static void
ScriptRunButton(GtkWidget *w, ScriptButtonType * sb)
{
    if (sb->type & SB_SCRIPT)
	ScriptRunFromString(sb->str);
    else if (sb->type & SB_SCRIPTFILE)
	ScriptRunFromFile(sb->str);
    else
	Term->remoteWrite(sb->str, strlen(sb->str));

    SoundPlay(SoundClickFile);

    w = NULL;	/* to make gcc happy */
}

/* ScriptRemoveBookmark() {{{1 */
static void
ScriptRemoveBookmark(void)
{
    GSList *slist;
    ScriptBookmarkType *sb;
    GtkWidget *w;
    gboolean globalFound;

    if (BookmarkList && !BookmarkRemoved)
    {
	BookmarkRemoved = TRUE;
	DisconnectFuncList = g_slist_remove(DisconnectFuncList,
					    ScriptRemoveBookmark);

	/* ϴ entry ְ, branch ִ  branch ش.
	 *  entry branch ؿ ִ 쿡 branch ֵ ڵ sub
	 * menu  ׷  츦 ؼ item 
	 * ִ .
	 */
	for (slist = BookmarkList; slist; slist = slist->next)
	{
	    sb = slist->data;
	    if ((sb->type & (SB_BRANCH | SB_GLOBAL)) == 0)
		/* global ƴϰ branch ƴϸ */
		gtk_item_factory_delete_item(MenuFactory, sb->id);
	}
	globalFound = FALSE;
	for (slist = BookmarkList; slist;)
	{
	    sb = slist->data;

	    /* list remove Ͼ   list   ϴ. */
	    slist = slist->next;

	    if (sb->type & SB_GLOBAL)
		globalFound = TRUE;
	    else
	    {
		if (sb->type & SB_BRANCH)
		{
		    /* global ƴϰ branch̸ */
		    w = gtk_item_factory_get_widget(MenuFactory, sb->id);
		    w = gtk_widget_get_ancestor(w, GTK_TYPE_MENU);
		    w = gtk_menu_get_attach_widget(GTK_MENU(w));
		    if (w)
			gtk_widget_destroy(w);
		    else
			g_warning("w = nil");
		}
		g_free(sb->id);
		g_free(sb->str);
		g_free(sb);

		BookmarkList = g_slist_remove(BookmarkList, sb);
	    }
	}
	if (!globalFound)
	{
	    g_slist_free(BookmarkList);
	    BookmarkList = NULL;
	}
    }
}

/* ScriptRemoveButton() {{{1 */
static void
ScriptRemoveButton(void)
{
    GSList *slist;
    ScriptButtonType *sb;
    gboolean globalFound;

    if (ButtonList && !ButtonRemoved)
    {
	ButtonRemoved = TRUE;
	DisconnectFuncList = g_slist_remove(DisconnectFuncList,
					    ScriptRemoveButton);

	globalFound = FALSE;
	for (slist = ButtonList; slist;)
	{
	    sb = slist->data;

	    /* list remove Ͼ   list   ϴ. */
	    slist = slist->next;

	    if (sb->type & SB_GLOBAL)
		globalFound = TRUE;
	    else
	    {
		gtk_widget_destroy(GTK_WIDGET(sb->id));
		g_free(sb->str);
		g_free(sb);

		ButtonList = g_slist_remove(ButtonList, sb);
	    }
	}
	if (!globalFound)
	{
	    g_slist_free(ButtonList);
	    ButtonList = NULL;
	}
    }
}

/* ScriptAssign() {{{1 */
static gint
ScriptAssign(ScriptType *script)
{
    return AddAssign(script->argv[0], script->argv[1]);
}

/* ScriptAutoRes() {{{1 */
static gint
ScriptAutoRes(ScriptType *script)
{
    /* argv[0]: add or gadd
     * argv[1]: ٸ ڿ
     * argv[2]: string, script
     * argv[3]: action string or script
     */
    if (!strcasecmp(script->argv[0], "add")
	|| !strcasecmp(script->argv[0], "gadd"))
    {
	AutoResponseAdd(script->argv[0], script->argv[1], script->argv[2],
			script->argv[3]);
	return OK;
    }
    else
	g_warning("Unknown token '%s' in autores.", (char *) script->argv[0]);
    return ERROR;
}

/* ScriptBookmark() {{{1 */
static gint
ScriptBookmark(ScriptType * script)
{
    char *p;
    int type;
    GtkItemFactoryEntry entry = { NULL, NULL, NULL, 0, _GN(NULL )};
    ScriptBookmarkType *sb;

    /* argv[0]: add
     * argv[1]: bookmark title
     * argv[2]: string, script
     * argv[3]: action string or script
     */
    if (!strcasecmp(script->argv[0], "add"))
	type = 0;
    else if (!strcasecmp(script->argv[0], "gadd"))
	type = SB_GLOBAL;
    else
	return ERROR;

    p = PC_IconvStr(script->argv[1], strlen(script->argv[1]), 256);
    entry.path = g_strconcat("/Bookmarks/", p, NULL);
    if (p != script->argv[1])
	g_free(p);

    /* script  script item destroyǹǷ ؼ
     * callback_data Ѱܾ ȴ. , ⿡ convert.
     */
    p = ConvertCtrlChar(script->argv[3]);

    sb = g_new(ScriptBookmarkType, 1);
    sb->type = type;
    sb->id = entry.path;
    if (!strcasecmp(script->argv[2], "script"))
	sb->type |= SB_SCRIPT;
    else if (!strcasecmp(script->argv[2], "sfile"))
	sb->type |= SB_SCRIPTFILE;
    sb->str = p;

    if (!strcasecmp(script->argv[3], "<BRANCH>"))
    {
	sb->type |= SB_BRANCH;
	entry.item_type = "<Branch>";
	entry.callback = NULL;
    }
    else
    {
	entry.item_type = "";
	entry.callback = ScriptRunBookmark;
	gtk_item_factory_create_item(MenuFactory, &entry, sb, 1);
    }

    if (BookmarkRemoved)
    {
	DisconnectFuncList = g_slist_append(DisconnectFuncList,
					    ScriptRemoveBookmark);
	BookmarkRemoved = FALSE;
    }

    BookmarkList = g_slist_prepend(BookmarkList, sb);

    return OK;
}

/* ScriptButton() {{{1 */
static gint
ScriptButton(ScriptType * script)
{
    char *p;
    int type;
    GtkWidget *button;
    ScriptButtonType *sb;

    /* argv[0]: add
     * argv[1]: bookmark title
     * argv[2]: string, script
     * argv[3]: action string or script
     */
    if (!strcasecmp(script->argv[0], "add"))
	type = 0;
    else if (!strcasecmp(script->argv[0], "gadd"))
	type = SB_GLOBAL;
    else
	return ERROR;

    if (!GTK_WIDGET_VISIBLE(CtrlBox))
	gtk_widget_show(CtrlBox);

    sb = g_new(ScriptButtonType, 1);

    p = PC_IconvStr(script->argv[1], strlen(script->argv[1]), 256);
    button = gtk_button_new_with_label(p);
    if (p != script->argv[1])
	g_free(p);

    gtk_button_set_relief(GTK_BUTTON(button), GTK_RELIEF_NONE);
    GTK_WIDGET_UNSET_FLAGS(button, GTK_CAN_FOCUS);
    gtk_box_pack_start(GTK_BOX(CtrlBox), button, FALSE, FALSE, 0);
    gtk_signal_connect(GTK_OBJECT(button), "clicked",
		       GTK_SIGNAL_FUNC(ScriptRunButton), sb);
    gtk_widget_show(button);

    sb->id = (char *) button;

    sb->type = type;
    if (!strcasecmp(script->argv[2], "script"))
	sb->type |= SB_SCRIPT;
    else if (!strcasecmp(script->argv[2], "sfile"))
	sb->type |= SB_SCRIPTFILE;

    p = ConvertCtrlChar(script->argv[3]);
    sb->str = p;

    if (ButtonRemoved)
    {
	DisconnectFuncList = g_slist_append(DisconnectFuncList,
					    ScriptRemoveButton);
	/* should be removed after disconnected */
	ButtonRemoved = FALSE;
    }
    ButtonList = g_slist_append(ButtonList, sb);
    ControlBarFix();

    return OK;
}

/* ScriptControlMenu() {{{1 */
static gint
ScriptControlMenu(ScriptType * script)
{
    gboolean global;

    if (!strcasecmp(script->argv[0], "gadd"))
	global = TRUE;
    else if (!strcasecmp(script->argv[0], "add"))
	global = FALSE;
    else
	return ERROR;

    ControlBarAddMenu(global, script->argv[1], script->argv[2],
		      script->argv[3], script->argv[4]);
    ControlBarFix();

    return OK;
}

/* ScriptRemoveLabel() {{{1 */
static void
ScriptRemoveLabel(void)
{
    GSList *slist;

    if (ScriptLabelList && !ScriptLabelRemoved)
    {
	for (slist = ScriptLabelList; slist; slist = slist->next)
	{
	    if (!strcasecmp(slist->data, "baudrate"))
		ControlLabelBaudCtrl(FALSE);
	    else if (!strcasecmp(slist->data, "emulate"))
		ControlLabelEmulCtrl(FALSE);
	    else if (!strcasecmp(slist->data, "protocol"))
		ControlLabelProtCtrl(FALSE);
	    else if (!strcasecmp(slist->data, "linestatus"))
		ControlLabelLineStsCtrl(FALSE);
	    else
		g_warning("%s: BUG! invalid label-list", MyName);
	    g_free(slist->data);
	}
	g_slist_free(ScriptLabelList);
	ScriptLabelList = NULL;
	ScriptLabelRemoved = TRUE;
    }
}

/* ScriptLabel() {{{1 */
static gint
ScriptLabel(ScriptType * script)
{
    gboolean globflag = FALSE, addflag = TRUE;
    int result;

    if (!strcasecmp(script->argv[0], "add"))
	globflag = FALSE;
    else if (!strcasecmp(script->argv[0], "gadd"))
	globflag = TRUE;
    else if (!strcasecmp(script->argv[0], "rmv"))
	addflag = FALSE;
    else
	return ERROR;

    if (!strcasecmp(script->argv[1], "baudrate"))
	result = ControlLabelBaudCtrl(addflag);
    else if (!strcasecmp(script->argv[1], "emulate"))
	result = ControlLabelEmulCtrl(addflag);
    else if (!strcasecmp(script->argv[1], "protocol"))
	result = ControlLabelProtCtrl(addflag);
    else if (!strcasecmp(script->argv[1], "linestatus"))
	result = ControlLabelLineStsCtrl(addflag);
    else
	result = ERROR;

    if (result == OK)
    {
	if (!globflag && addflag)
	{
	    ScriptLabelList = g_slist_append(ScriptLabelList,
					     g_strdup(script->argv[1]));
	    if (ScriptLabelRemoved)
	    {
		DisconnectFuncList = g_slist_append(DisconnectFuncList,
						    ScriptRemoveLabel);
		ScriptLabelRemoved = FALSE;
	    }
	}
	ControlBarFix();
    }
    return result;
}

/* ScriptIf() {{{1 */
static int
ScriptIf(ScriptType *script)
{
    int sts;
    GList *list, *prev;
    ScriptType *scr;
    gint depth, v;
    gboolean same;
    char **argv;

    if (script->argc != 3)
	return ERROR;

    argv = (char **) script->argv;

    if (!strcmp(argv[1], "=="))
	same = TRUE;
    else if (!strcmp(argv[1], "!="))
	same = FALSE;
    else
	return ERROR;

    sts = 0;
    if (!strcmp(argv[0], "$?"))
    {
	/*    Ȯ */
	v = atoi(argv[2]);
	if ((same && ScriptResult == v) || (!same && ScriptResult != v))
	    sts = 0;
	else
	    sts = -1;
    }
    else
    {
	int c;

	c = strcmp(argv[0], argv[2]);
	if ((same && !c) || (!same && c))
	    sts = 0;
	else
	    sts = -1;
    }

    if (sts < 0)
    {
	/*    ENDIF ã´. */
	depth = 0;
	prev = NULL;
	for (list = CurrScriptItem->next; list; list = list->next)
	{
	    scr = (ScriptType *) list->data;
	    if (scr->type == SCR_IF)
		++depth;
	    else if (scr->type == SCR_ENDIF || scr->type == SCR_ELSE
		     || scr->type == SCR_ELSEIF)
	    {
		if (depth <= 0)
		{
		    if (scr->type == SCR_ELSEIF)
			/* CurrScriptItem prev Ͽ elseif
			 * ǵ Ѵ.
			 */
			CurrScriptItem = prev;
		    else
			/* CurrScriptItem list Ͽ   script
			 * ǵ Ѵ.
			 */
			CurrScriptItem = list;
		    break;
		}
		if (scr->type == SCR_ENDIF)
		    --depth;
	    }
	    prev = list;
	}
	if (!list)
	{
	    ScriptFinish();
	    StatusShowMessage(_("CANNOT FIND MATCHING ENDIF!"));
	}
    }
    return OK;
}

/* ScriptElse() {{{1 */
static int
ScriptElse(ScriptType * script)
{
    GList *list;
    ScriptType *scr;
    gint depth;

    depth = 0;
    for (list = CurrScriptItem->next; list; list = list->next)
    {
	scr = list->data;
	if (scr->type == SCR_IF)
	    ++depth;
	else if (scr->type == SCR_ENDIF)
	{
	    if (depth <= 0)
	    {
		CurrScriptItem = list;
		break;
	    }
	    --depth;
	}
    }
    if (!list)
    {
	ScriptFinish();
	StatusShowMessage(_("CANNOT FIND MATCHING ENDIF!"));
    }
    return OK;

    script = NULL;
}

/* ScriptRunSignalHandler() {{{1 */
static void
ScriptRunSignalHandler(void)
{
    if (CurrScriptItem && CurrScriptItem->next)
    {
	CurrScriptItem = CurrScriptItem->next;
	ScriptInstallNextItem();
    }
    else
	ScriptFinish();
    gdk_input_remove(ScriptRunInputTag);
    close(ScriptRunFD[0]);
    TermReadInputTag = gdk_input_add(ReadFD, GDK_INPUT_READ,
				     TermInputHandler, NULL);

    if (ScriptChildStr)
    {
	ScriptRunFromString(ScriptChildStr);
	g_free(ScriptChildStr);
	ScriptChildStr = NULL;
    }
}

/* ScriptRunQueueString() {{{1 */
static void
ScriptRunQueueString(char *buf)
{
    char *s;

    if (ScriptChildStr == NULL)
	ScriptChildStr = g_strdup("");
    if (ScriptChildStr[strlen(ScriptChildStr) - 1] != ';')
    {
	s = ScriptChildStr;
	ScriptChildStr = g_strconcat(s, buf, ";", NULL);
	g_free(s);
    }
}

/* ScriptRunInputHandler() {{{1 */
static void
ScriptRunInputHandler(gpointer data, int source, GdkInputCondition cond)
{
    char *p, buf[256];
    int i, size;

    if ((size = read(source, buf, sizeof(buf) - 1)) > 0)
    {
	buf[size] = '\0';
	if (!strncasecmp(buf, "/set script ", 12))
	{
	    for (p = buf; *p; p++)
	    {
		if (*p == '\'' || *p == '"')
		{
		    char c = *p++;
		    while (*p && *p != c)
			p++;
		}
		if (*p == '\n' || *p == '\r')
		    *p = ';';
	    }
	    p = buf + 12;
	    while ((p = strstr(p, "/set script ")) != NULL)
	    {
		for (i = 0; i < 12; i++)
		    *p++ = ' ';
	    }
	    ScriptRunQueueString(buf + 12);
	}
	else
	    Term->localWrite(Term, buf, size);
    }
    data = NULL;
    cond = 0;
}

/* ScriptRun() {{{1 */
static int
ScriptRun(ScriptType * script)
{
    pid_t pid;
    char *file, buf[MAX_PATH];

    /* GetRealFilename() read  checkѴ. , ุ
     *   ⿡ õȴ.
     */
    if ((file = GetRealFilename(script->argv[0], ScriptPath)) == NULL)
	return ERROR;
    strncpy(buf, file, sizeof(buf) - 1);
    g_free(file);

    if (access(buf, X_OK))
    {
	g_warning("'%s' cannot be executed", buf);
	return ERROR;
    }

    if (pipe(ScriptRunFD) == -1)
    {
	g_warning("%s: Cannot create pipe", __FUNCTION__);
	return ERROR;
    }

    if ((pid = fork()) == -1)
    {
	g_warning("%s: Cannot create child process", __FUNCTION__);
	return ERROR;
    }

    if (pid > 0)
    {
	/* I'm parent */
	AddChildSignalHandler(pid, ScriptRunSignalHandler);
	gdk_input_remove(TermReadInputTag);
	TermReadInputTag = 0;
	close(ScriptRunFD[1]);
	ScriptRunInputTag = gdk_input_add(ScriptRunFD[0], GDK_INPUT_READ,
					  ScriptRunInputHandler, NULL);
	return OK;
    }

    /* I'm child */
    dup2(ReadFD, 0);
    dup2(ReadFD, 1);
    close(ReadFD);
    close(ScriptRunFD[0]);
    dup2(ScriptRunFD[1], 2);
    execlp("sh", "sh", "-c", buf, NULL);

    /* should never reach if command execute success */
    g_warning("%s: Fail to run '%s'", __FUNCTION__, buf);
    _exit(1);
}

/* ScriptDestroyWinObjList() {{{1 */
static void
ScriptDestroyWinObjList(void)
{
    GSList *slist;
    ScriptWinObjType *obj;

    if (ScriptWinObjList)
    {
	for (slist = ScriptWinObjList; slist; slist = slist->next)
	{
	    obj = slist->data;
	    g_free(obj->str);
	    if (obj->env)
		g_free(obj->env);
	    g_free(obj);
	}
	g_slist_free(ScriptWinObjList);
	ScriptWinObjList = NULL;
    }
}

/* ScriptWinDestroy() {{{1 */
static void
ScriptWinDestroy(GtkWidget **win)
{
    GtkWidget *w;
    w = *win;

    /* quit button  ̿ϴ   gtk_widget_destroy() 
     * win destroy signal ޾Ƽ  Լ ٽ ҷ. ,
     * quit button ̿ϸ  Լ ι ҷǷ w NULL
     * Ȯؾ Ѵ.
     */
    if (w)
    {
	*win = NULL;
	gtk_widget_destroy(w);
    }

    ScriptDestroyWinObjList();

    if (ScriptWinSavedList)
    {
	CurrScriptItem = ScriptWinSavedList;
	ScriptWinSavedList = NULL;
	ScriptInstallNextItem();
    }
}

/* ScriptWinButtonClicked() {{{1 */
static void
ScriptWinButtonClicked(GtkWidget *w, gpointer data)
{
    char *p, *p2, *p3, *str, *scriptStr;
    const char *entryStr;
    char *quote;
    ScriptWinObjType *obj;

    obj = gtk_object_get_user_data(GTK_OBJECT(w));
    str = g_strdup(obj->str);
    if (data)
    {
	entryStr = gtk_entry_get_text(GTK_ENTRY(data));
	if (!*entryStr)
	    return;

	if ((p = strstr(str, obj->env)) != NULL)
	{
	    /* '$' keyword ' " ̿   ֵǥ ´. */
	    if (strpbrk(entryStr, " \t"))
	    {
		quote = "\"";
		if ((p2 = strpbrk(str, "'\"")) != NULL)
		{
		    if ((p3 = strchr(p2 + 1, *p2)) != NULL)
		    {
			if (p2 < p && p < p3)
			    quote = "";
		    }
		}
	    }
	    else
		quote = "";

	    *p = '\0';
	    p2 = strpbrk(p + strlen(obj->env), " \t\n\r\\;'\"");
	    scriptStr = g_strconcat(str, quote, entryStr, quote, p2, NULL);
	    g_free(str);
	}
	else
	{
	    g_warning("%s: internal ERROR", __FUNCTION__);
	    scriptStr = str;
	}
	gtk_entry_set_text(GTK_ENTRY(data), "");
    }
    else
	scriptStr = str;

    if (obj->type & SB_SCRIPT)
	ScriptRunFromString(scriptStr);
    else
    {
	str = ConvertCtrlChar(scriptStr);
	Term->remoteWrite(str, strlen(str));
	g_free(str);
    }

    g_free(scriptStr);
}

/* ScriptWinCreateVBox() {{{1 */
static GtkWidget *
ScriptWinCreateVBox(GtkWidget *w, char *str)
{
    int spacing;
    GtkWidget *vbox;

    if (w)
    {
	if (!str || sscanf(str, "%d", &spacing) != 1)
	    spacing = 3;
	vbox = gtk_vbox_new(FALSE, spacing);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 2);
	gtk_container_add(GTK_CONTAINER(w), vbox);
	gtk_widget_show(vbox);
	return vbox;
    }
    else
	g_warning("Cannot create vbox. Parent widget is null");
    return NULL;
}

/* ScriptWinCreateHBox() {{{1 */
static GtkWidget *
ScriptWinCreateHBox(GtkWidget *w, char *str)
{
    int spacing;
    GtkWidget *hbox;

    if (w)
    {
	if (!str || sscanf(str, "%d", &spacing) != 1)
	    spacing = 3;
	hbox = gtk_hbox_new(FALSE, spacing);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 2);
	gtk_container_add(GTK_CONTAINER(w), hbox);
	gtk_widget_show(hbox);
	return hbox;
    }
    else
	g_warning("Cannot create hbox. Parent widget is null");
    return NULL;
}

typedef struct {
    char *name;
    GtkWidget *w;
} ScriptParentWidgetType;

/* ScriptFindParentWidget() {{{1 */
static GtkWidget *
ScriptFindParentWidget(GSList *slist, char *name)
{
    ScriptParentWidgetType *spw;

    for (; slist; slist = slist->next)
    {
	spw = slist->data;
	if (!strcmp(name, spw->name))
	    return spw->w;
    }
    return NULL;
}

/* ScriptDestroyParentWidgetSList() {{{1 */
static void
ScriptDestroyParentWidgetSList(GSList *slist)
{
    if (slist)
    {
	for (; slist; slist = slist->next)
	    g_free(slist->data);
	g_slist_free(slist);
    }
}

/* ScriptGenSPW() {{{1 */
static gpointer
ScriptGenSPW(GtkWidget *w, char *name)
{
    ScriptParentWidgetType *spw;

    spw = g_new(ScriptParentWidgetType, 1);
    spw->w = w;

    /* dup ʿ. ScriptCreateWindow()  Ǳ ̴. */
    spw->name = name;
    return spw;
}

/* ScriptCreateWindow() {{{1 */
static int
ScriptCreateWindow(void)
{
    register int i;
    ScriptType *script;
    GtkWidget *w, *hbox, *vbox, *entry, *hsep, *button, *label, *frame;
#define MAX_ENTRY_NUM 2
    int entryNum;
    struct {
	char *str;
	char *actButton;
	GtkWidget *entry;
    } entryTbl[MAX_ENTRY_NUM];
    char *p;
    GSList *slist;

    if (ScriptWin)
    {
	if (GTK_WIDGET_VISIBLE(ScriptWin))
	    gtk_widget_destroy(ScriptWin);
	ScriptWin = NULL;
    }
    w = hbox = vbox = NULL;
    slist = NULL;
    entryNum = 0;
    for (;;)
    {
	script = CurrScriptItem->data;

	switch (script->type)
	{
	    case SCR_WIN_WINDOW:
		ScriptWin = gtk_window_new(GTK_WINDOW_TOPLEVEL);
		gtk_window_set_position(GTK_WINDOW(ScriptWin),
					GTK_WIN_POS_MOUSE);
		if (script->argc >= 1)
		    gtk_window_set_title(GTK_WINDOW(ScriptWin),
					 script->argv[0]);
		else
		    gtk_window_set_title(GTK_WINDOW(ScriptWin), "Script Win");
		gtk_window_set_policy(GTK_WINDOW(ScriptWin), TRUE, TRUE, FALSE);
		gtk_signal_connect_object(GTK_OBJECT(ScriptWin), "destroy",
					  (GtkSignalFunc) ScriptWinDestroy,
					  (GtkObject *) &ScriptWin);
		slist = g_slist_append(slist, ScriptGenSPW(ScriptWin, "win"));
		break;
	    case SCR_WIN_VBOX:
		w = ScriptFindParentWidget(slist, script->argv[0]);
		if (!w)
		    goto err_return;
		vbox = ScriptWinCreateVBox(w, script->argv[2]);
		if (vbox == NULL)
		    goto err_return;
		slist = g_slist_append(slist,
				       ScriptGenSPW(vbox, script->argv[1]));
		break;
	    case SCR_WIN_HBOX:
		w = ScriptFindParentWidget(slist, script->argv[0]);
		if (!w
		    || (hbox = ScriptWinCreateHBox(w, script->argv[2])) == NULL)
		    goto err_return;
		slist =
		    g_slist_append(slist, ScriptGenSPW(hbox, script->argv[1]));
		break;

	    case SCR_WIN_FRAME:
		w = ScriptFindParentWidget(slist, script->argv[0]);
		if (!w)
		    goto err_return;
		frame = gtk_frame_new((char *) script->argv[2]);
		gtk_container_add(GTK_CONTAINER(w), frame);
		gtk_widget_show(frame);
		slist = g_slist_append(slist,
				       ScriptGenSPW(frame, script->argv[1]));
		break;

	    case SCR_WIN_LABEL:
		w = ScriptFindParentWidget(slist, script->argv[0]);
		if (!w)
		    goto err_return;
		label = gtk_label_new((char *) script->argv[1]);
		gtk_container_add(GTK_CONTAINER(w), label);
		gtk_widget_show(label);
		break;

	    case SCR_WIN_ENTRY:
		w = ScriptFindParentWidget(slist, script->argv[0]);
		if (!w)
		    goto err_return;

		if (entryNum < MAX_ENTRY_NUM
		    && ((char **) script->argv)[1][0] == '$')
		{
		    entry = gtk_entry_new();
		    gtk_box_pack_start(GTK_BOX(w), entry, TRUE, TRUE, 0);
		    gtk_widget_show(entry);

		    entryTbl[entryNum].str = script->argv[1];
		    entryTbl[entryNum].entry = entry;
		    if (script->argc >= 3
			&& !strcasecmp(script->argv[2], "act"))
			entryTbl[entryNum].actButton = script->argv[3];
		    else
			entryTbl[entryNum].actButton = NULL;

		    ++entryNum;
		}
		else
		{
		    g_warning("entry error. num=%d, argv[1][0]=%c",
			      entryNum, ((char **) script->argv)[1][0]);
		    goto err_return;
		}
		break;

	    case SCR_WIN_BUTTON:
		w = ScriptFindParentWidget(slist, script->argv[0]);
		if (!w)
		    goto err_return;

		button = gtk_button_new_with_label(script->argv[1]);
		gtk_box_pack_start(GTK_BOX(w), button, TRUE, TRUE, 0);
		if (!strcasecmp(script->argv[2], "quit"))
		    gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
					      (GtkSignalFunc) ScriptWinDestroy,
					      (GtkObject *) &ScriptWin);
		else
		{
		    gpointer data;
		    ScriptWinObjType *obj;

		    obj = g_new(ScriptWinObjType, 1);
		    obj->str = g_strdup(script->argv[3]);
		    obj->env = NULL;

		    if (!strcasecmp(script->argv[2], "script"))
			obj->type = SB_SCRIPT;
		    else if (!strcasecmp(script->argv[2], "string"))
			obj->type = 0;	/* string */
		    else
		    {
			g_free(obj->str);
			g_free(obj);
			goto err_return;
		    }

		    data = NULL;

		    for (i = 0; i < entryNum; i++)
		    {
			if ((p = strstr(script->argv[3], entryTbl[i].str)) !=
			    NULL)
			{
			    p += strlen(entryTbl[i].str);
			    if (!*p || isspace(*p) || *p == '\\' || *p == ';')
			    {
				data = entryTbl[i].entry;
				obj->env = g_strdup(entryTbl[i].str);

				if (entryTbl[i].actButton
				    && !strcmp(script->argv[1],
					       entryTbl[i].actButton))
				{
				    /*  ư active button ߴٸ entry
				     *  object ѱ, activate signal 
				     */
				    gtk_signal_connect(GTK_OBJECT(data),
						       "activate",
						       (GtkSignalFunc)ScriptWinButtonClicked,
						       data);
				    gtk_object_set_user_data(GTK_OBJECT(data),
							     obj);
				}
				break;
			    }
			}
		    }

		    /* list ؼ ߿  free */
		    ScriptWinObjList = g_slist_append(ScriptWinObjList, obj);
		    gtk_object_set_user_data(GTK_OBJECT(button), obj);
		    gtk_signal_connect(GTK_OBJECT(button), "clicked",
				       (GtkSignalFunc)ScriptWinButtonClicked,
				       data);
		}
		gtk_widget_show(button);
		break;

	    case SCR_WIN_HSEP:
		w = ScriptFindParentWidget(slist, script->argv[0]);
		if (w)
		{
		    hsep = gtk_hseparator_new();
		    gtk_box_pack_start(GTK_BOX(w), hsep, FALSE, FALSE, 3);
		    gtk_widget_show(hsep);
		}
		else
		    g_warning("Cannot create hseperator. vbox is null");
		/* 뼼   Ƿ   */
		break;

	    case SCR_WIN_WAIT:
	    case SCR_WIN_END:
		if (ScriptWin)
		    gtk_widget_show(ScriptWin);
		ScriptDestroyParentWidgetSList(slist);
		return OK;

	    default:
err_return:
		/* help debug */
		g_print("ERROR: maybe you make a invalid script arguments.\n"
			"ARGS:");
		for (i = 0; i < script->argc; i++)
		    g_print(" %s", (char *) script->argv[i]);
		g_print("\n");

		CurrScriptItem->next = NULL;	/* finish the script */
		break;
	}

	CurrScriptItem = CurrScriptItem->next;
	if (CurrScriptItem == NULL)
	    break;
    }

    if (ScriptWin)
    {
	gtk_widget_destroy(ScriptWin);
	ScriptWin = NULL;
    }

    ScriptDestroyParentWidgetSList(slist);
    ScriptDestroyWinObjList();
    ScriptFinish();
    return ERROR;
}

/* ScriptAssignConvert() {{{1 */
static void
ScriptAssignConvert(ScriptType *script)
{
    int i;
    AssignType *a = NULL;
    char *s, *e, *f;
    char *sa;

    if (script->savedArgv)
    {
	for (i = 0; i < script->argc; i++)
	{
	    if (script->savedArgv[i])
	    {
		g_free(script->argv[i]);
		script->argv[i] = g_strdup(script->savedArgv[i]);
	    }
	}
    }

    for (i = 0; i < script->argc; i++)
    {
	sa = script->argv[i];
	while (*sa && (s = strchr(sa, '$')) != NULL)
	{
	    if (s != sa && s[-1] == '\\')
	    {
		sa = s + 1;
		continue;
	    }
	    if ((e = FindWordSeperator(s + 1)) == NULL)
		f = g_strdup(s + 1);
	    else
	    {
		guint l = (guint) (e - s - 1);

		if (l)
		{
		    f = g_malloc(l + 1);
		    memcpy(f, s + 1, l);
		    f[l] = '\0';
		}
		else
		{
		    sa = s + 1;
		    continue;
		}
	    }

	    a = GetAssign(f);
	    g_free(f);

	    if (script->savedArgv == NULL)
		script->savedArgv = g_new0(gpointer, script->argc);
	    if (script->savedArgv[i] == NULL)
		script->savedArgv[i] = g_strdup(script->argv[i]);

	    sa = script->argv[i];
	    if (a == NULL)
		script->argv[i] = g_strdup("");
	    else
	    {
		*s = '\0';
		script->argv[i] = g_strdup_printf("%s%s%s", *sa? sa: "",
						  a->val,
						  e? e: "");
	    }
	    g_free(sa);
	    sa = script->argv[i];
	}
    }
}

/* ScriptRunNext() {{{1 */
static int
ScriptRunNext(void)
{
    ScriptType *script;
    unsigned long ul;
    gchar *p;
    gboolean doInstallNext;

    if (CurrScriptItem == NULL)
    {
	ScriptFinish();
	return FALSE;
    }

    IdleTag = 0;
    PauseTag = 0;
    doInstallNext = TRUE;

    script = CurrScriptItem->data;
    ScriptAssignConvert(script);

    switch (script->type)
    {
	case SCR_ASSIGN:
	    ScriptResult = ScriptAssign(script);
	    break;
	case SCR_AUTORES:
	    ScriptResult = ScriptAutoRes(script);
	    break;
	case SCR_BEEP:
	    if (SoundPlay(SoundBeepFile) != OK)
	    {
		int percent = (int) BeepPercent;
		if (script->argc > 0)
		{
		    if (sscanf(script->argv[0], "%d", &percent) != 1
			|| percent == 0)
			percent = (int) BeepPercent;
		}
		XBell(gdk_display, percent);
	    }
	    ScriptResult = OK;
	    break;
	case SCR_BOOKMARK:
	    ScriptResult = ScriptBookmark(script);
	    break;
	case SCR_BUTTON:
	    ScriptResult = ScriptButton(script);
	    break;
	case SCR_CAPTURE:
	    if (!strcasecmp(script->argv[0], "on"))
	    {
		if (script->argc >= 2)
		{
		    gboolean includeBuf;

		    ScriptCaptureActive = TRUE;
		    if (script->argc == 3)
			includeBuf = !strcasecmp(script->argv[2], "includebuf");
		    else
			includeBuf = FALSE;
		    ScriptResult = CaptureStart(script->argv[1], includeBuf);
		    break;
		}
	    }
	    else if (!strcasecmp(script->argv[0], "off"))
	    {
		ScriptCaptureActive = FALSE;
		CaptureFinish();
		ScriptResult = OK;
		break;
	    }
	    ScriptResult = ERROR;
	    break;
	case SCR_CLEAR:
	    TermClearScreen(Term);
	    ScriptResult = OK;
	    break;
	case SCR_CTLMENU:
	    ScriptResult = ScriptControlMenu(script);
	    break;
	case SCR_DISCONNECT:
	    PC_Hangup();
	    ScriptResult = OK;
	    break;
	case SCR_DOWNLOAD:
	    if (!strcasecmp(script->argv[0], "zmodem"))
	    {
		TRxProtocolSetFromString(script->argv[0]);
		RunRZ();
		ScriptResult = OK;
		break;
	    }
	    /* TODO: zmodem ƴ   */
	    ScriptResult = ERROR;
	    break;
	case SCR_ELSE:
	    ScriptResult = ScriptElse(script);
	    break;
	case SCR_EMULATE:
	    if (!strcasecmp(script->argv[0], "ANSI"))
		EmulateMode = EMULATE_ANSI;
	    else if (!strcasecmp(script->argv[0], "VT100"))
		EmulateMode = EMULATE_VT100;
	    else
	    {
		ScriptResult = ERROR;
		break;
	    }
	    TermEmulModeSet();
	    break;
	case SCR_END:
	    ScriptFinish();
	    break;
	case SCR_ENDIF:
	    /* do nothing */
	    ScriptResult = OK;
	    break;
	case SCR_ENDWIN:
	    if (ScriptWin)
	    {
		gtk_widget_destroy(ScriptWin);
		ScriptWin = NULL;
	    }
	    break;
	case SCR_GET_BAUDRATE:
	    ScriptResult = ScriptGetBaudrate(script);
	    break;
	case SCR_GET_FILEINFO:
	    ScriptResult = ScriptGetFileInfo(script);
	    break;
	case SCR_GOTO:
	    ScriptResult = ScriptGoto(script->argv[0]);
	    break;
	case SCR_GOTO_TAG:
	    /* do nothing */
	    ScriptResult = OK;
	    break;
	case SCR_IF:
	case SCR_ELSEIF:
	    ScriptResult = ScriptIf(script);
	    break;
	case SCR_KFLUSH:
	    /* procomm plus script ȣȯ */
	    ScriptResult = OK;
	    break;
	case SCR_LABEL:
	    ScriptResult = ScriptLabel(script);
	    break;
	case SCR_MESSAGE:
	    p = ConvertCtrlChar(script->argv[0]);
	    Term->localWrite(Term, p, strlen(p));
	    g_free(p);
	    ScriptResult = OK;
	    break;
	case SCR_PAUSE:
	    /* pause ̸ pause ǹ̰ ְڴ°? */
	    if (CurrScriptItem->next)
	    {
		float f;
		if (sscanf(script->argv[0], "%f", &f) == 1 && f >= 0.1)
		    PauseTag = gtk_timeout_add((guint32) (f * 1000UL),
					       (GtkFunction) ScriptRunNext,
					       NULL);
		else
		    script->type = SCR_NONE;
	    }
	    ScriptResult = OK;
	    break;
	case SCR_RFLUSH:
	    /* procomm plus script ȣȯ */
	    ScriptResult = OK;
	    break;
	case SCR_RUN:
	    ScriptResult = ScriptRun(script);
	    doInstallNext = FALSE;
	    break;
	case SCR_TRANSMIT:
	    p = ConvertCtrlChar(script->argv[0]);
	    Term->remoteWrite(p, strlen(p));
	    g_free(p);
	    ScriptResult = OK;
	    break;
	case SCR_UPLOAD:
	    TRxProtocolSetFromString(script->argv[0]);
	    ScriptResult = OK;
	    if (*(char *) script->argv[1] == '?')
		UploadFileQuery();
	    else
	    {
		char *txfname = g_strdup(script->argv[1]);

		if (TRxSendFile(txfname) < 0)
		{
		    g_warning("Cannot upload '%s'.", txfname);
		    ScriptResult = ERROR;
		}
		g_free(txfname);
	    }
	    while (TRxRunning != TRX_IDLE)
		PC_Sleep(500);
	    break;
	case SCR_WAITFOR:
	    /* waitfor ̸ waitfor ǹ̰ ְڴ°? */
	    doInstallNext = FALSE;
	    if (CurrScriptItem->next)
	    {
		float waitTime = 0;
		if (script->argc >= 2)
		    if (sscanf(script->argv[1], "%f", &waitTime) != 1)
			waitTime = 0;

		ScriptWaitFor(script, waitTime);
	    }
	    break;

	case SCR_SET_BAUDRATE:
	    if (sscanf(script->argv[0], "%lu", &ul) != 1
		|| ModemSetSpeed(ul) == FALSE)
	    {
		StatusShowMessage(_("Cannot set baudrate to '%lu'."),
				  (unsigned long) ul);
		ScriptResult = ERROR;
	    }
	    else
	    {
		ScriptResult = OK;
		doInstallNext = FALSE;	/* window  next item install */
		CurrScriptItem = CurrScriptItem->next;
		/* baudrate ٲ Ŀ 2  device ų ʵ
		 * Ѵ. ̰  ġ. 2 ̳  Ÿ ׳
		 * .
		 */
		PauseTag = gtk_timeout_add(2500UL,
					   (GtkFunction) ScriptRunNext, NULL);
	    }
	    break;
	case SCR_SET_CONFIG:
	    ScriptResult =
		ConfigVarChange(script->argv[0], script->argv[1], TRUE);
	    break;
	case SCR_SET_DEBUG:
#if defined(_DEBUG_)
	    if (sscanf(script->argv[0], "%lu", &ul) == 1
		&& ul < ARYSIZ(OPrintFlag))
	    {
		OPrintFlag[ul] = !OPrintFlag[ul];
		ScriptResult = OK;
	    }
	    else
		ScriptResult = ERROR;
#endif /* _DEBUG_ */
	    break;
	case SCR_SET_PROTOCOL:
	    ScriptResult = TRxProtocolSetFromString(script->argv[0]);
	    break;
	case SCR_SET_TITLE:
	    SetWindowTitle(script->argv[0]);
	    ScriptResult = OK;
	    break;

	case SCR_WIN_WINDOW:
	    ScriptResult = ScriptCreateWindow();
	    if (CurrScriptItem)
	    {
		if (((ScriptType *) CurrScriptItem->data)->type == SCR_WIN_WAIT)
		{
		    doInstallNext = FALSE;	/* window  next item install */
		    ScriptWinSavedList = CurrScriptItem;
		}
	    }
	    break;
	case SCR_WIN_END:
	case SCR_WIN_WAIT:
	    /* nothing to do */
	    break;

	default:
	    CurrScriptItem = NULL;	/*  Ų. */
	    g_warning("BUG? Script finished with error type %d.", script->type);
	    break;
    }

    if (CurrScriptItem && CurrScriptItem->next)
    {
	if (script->type == SCR_PAUSE)
	    CurrScriptItem = CurrScriptItem->next;
	else if (doInstallNext)
	{
	    CurrScriptItem = CurrScriptItem->next;
	    script = CurrScriptItem->data;
	    if (script->type == SCR_WAITFOR)
	    {
		float waitTime = 0;

		ScriptAssignConvert(script);
		if (script->argc >= 2)
		{
		    if (sscanf(script->argv[1], "%f", &waitTime) != 1)
			waitTime = 0;
		}

		ScriptWaitFor(script, waitTime);
		return FALSE;
	    }
	    ScriptInstallNextItem();
	}
    }
    else
	ScriptFinish();

    return FALSE;
}

/* ScriptCancel() {{{1 */
void
ScriptCancel(void)
{
    if (CurrScriptFile)
	StatusShowMessage(_("Script '%s' cancelled."), CurrScriptFile);
    ScriptRemoveAll();
}

/* GetAssign() {{{1 */
static AssignType *
GetAssign(const char *s)
{
    GSList *slist;
    AssignType *assign;

    if (s && *s)
    {
	for (slist = AssignList; slist; slist = slist->next)
	{
	    assign = slist->data;
	    if (!strcmp(assign->var, s))
		return assign;
	}
    }
    return NULL;
}

/* ScriptParsingError() {{{1 */
static void
ScriptParsingError(char *param[], int nParam)
{
    int i;

    /* print as many as possible to make debug easy */
    g_print("%s: Invalid line in script. nParam=%d\n\t%s", __FUNCTION__,
	    nParam, param[0]);

    for (i = 1; i < nParam; i++)
	g_print(" %s", param[i]);
    g_print("\n");
}

/* AddAssign() {{{1 */
static int
AddAssign(char *var, char *val)
{
    AssignType *a;

    if ((a = GetAssign(var)) == NULL)
    {
	a = g_new(AssignType, 1);
	a->var = g_strdup(var);
	a->val = g_strdup(val);
	AssignList = g_slist_append(AssignList, a);
    }
    else
    {
	g_free(a->val);
	a->val = g_strdup(val);
    }
    return OK;
}

/* ScriptCreateItem() {{{1 */
static ScriptType *
ScriptCreateItem(char *param[], int nArg, int paramPos,
		 const ScriptTableType *tbl, int *result)
{
    register int i;
    ScriptType *script;
    char *p;

    *result = ERROR;	/* assume error */

    if (paramPos >= MAX_SCRIPT_PARAMS || nArg < 0)
    {
	ScriptParsingError(param, nArg + paramPos + 1);
	return NULL;
    }

    /* find the token from the table */
    for (; tbl->type >= 0; tbl++)
	if (!strcasecmp(param[paramPos], tbl->token))
	    break;

    if (tbl->type >= 0)
    {
	if (tbl->next)
	    return ScriptCreateItem(param, nArg - 1, paramPos + 1, tbl->next,
				    result);
	else
	{
	    script = g_new(ScriptType, 1);
	    script->type = tbl->type;
	    script->savedArgv = NULL;
	    if (nArg > 0 || tbl->argc > 0)
	    {
		if (nArg < tbl->argc)
		{
		    ScriptParsingError(param, nArg + paramPos + 1);
		    g_free(script);
		    return NULL;
		}
		script->argc = nArg;
		script->argv = g_new(gpointer, nArg);
		++paramPos;
		for (i = 0; i < nArg; i++)
		    script->argv[i] = g_strdup(param[paramPos + i]);
	    }
	    else
		script->argc = 0;
	    *result = OK;
	    return script;
	}
    }
    else
    {
	if (paramPos == 0)
	{
	    /* check if it is label */
	    p = param[0] + strlen(param[0]) - 1;
	    if (*p == ':')
	    {
		*p = '\0';
		script = g_new(ScriptType, 1);
		script->type = SCR_GOTO_TAG;
		script->argc = 1;
		script->argv = g_new(gpointer, 1);
		script->argv[0] = g_strdup(param[0]);
		script->savedArgv = NULL;
		*result = OK;
		return script;
	    }
	}
	ScriptParsingError(param, nArg + paramPos + 1);
    }
    return NULL;
}

/* ScriptInclude() {{{1 */
static GList *
ScriptInclude(GList *list, ScriptType * script)
{
    char *filename;
    GList *subList, *last;

    if ((filename = GetRealFilename(script->argv[0], ScriptPath)) != NULL)
    {
	subList = ScriptMakeListFromFile(filename);
	g_free(filename);

	if (subList != NULL)
	{
	    if (list == NULL)
		list = subList;
	    else
	    {
		last = g_list_last(list);
		last->next = subList;
		subList->prev = last;
	    }
	}
    }
    g_free(script->argv[0]);
    g_free(script->argv);
    g_free(script);
    return list;
}

/* ScriptInitCurrScriptItem() {{{1 */
static void
ScriptInitCurrScriptItem(GList *list)
{
    if (!ScriptItemList)
	CurrScriptItem = ScriptItemList = list;
    else
    {
	GList *last;

	CurrScriptItem->prev->next = list;
	list->prev = CurrScriptItem->prev;
	last = g_list_last(list);
	last->next = CurrScriptItem;
	CurrScriptItem->prev = last;
	CurrScriptItem = list;
    }
}

/* ScriptMakeListFromFile() {{{1 */
static GList *
ScriptMakeListFromFile(char *filename)
{
    int n, result;
    FILE *fp;
    GList *list;
    ScriptType *script;
    char *param[MAX_SCRIPT_PARAMS], *buf;

    if (filename == NULL || *filename == '\0')
    {
	StatusShowMessage(_("%s: filename is empty"), __FUNCTION__);
	return NULL;
    }

    if ((fp = fopen(filename, "r")) != NULL)
    {
	list = NULL;
	while ((buf = GetLines(fp)) != NULL)
	{
	    OptPrintf(OPRT_SCRIPT, "%s: s=%s\n", __FUNCTION__, buf);

	    if ((n = ParseLine(buf, MAX_SCRIPT_PARAMS, param)) < 1)
		continue;

	    OptPrintf(OPRT_SCRIPT, "%s: n=%d, param[0]=%s\n", __FUNCTION__,
		      n, param[0]);

	    script = ScriptCreateItem(param, n - 1, 0, ScriptMainTable,
				      &result);
	    if (script == NULL)
	    {
		if (result == OK)
		    /* just ignore this line! */
		    continue;

		g_list_free(list);
		list = NULL;
		break;
	    }

	    if (script->type == SCR_INCLUDE)
		list = ScriptInclude(list, script);
	    else
		list = g_list_append(list, script);
	}
	fclose(fp);
	return list;
    }
    else
	StatusShowMessage(_("%s: cannot open file '%s'"), __FUNCTION__,
			  filename);

    return NULL;
}

/* ScriptGetFullPath() {{{1 */
char *
ScriptGetFullPath(const char *filename)
{
    char *s;

    /* ϴ global script pathκ filename ִ  ã, 
     * filename  ã´.  code  filename absolute
     * path  ˻ ʴ´. ̰  ɱ?
     */
    if (strchr(filename, '.') == NULL)
    {
	char *buf = g_malloc(MAX_PATH);
	g_snprintf(buf, MAX_PATH, "%s" DEFAULT_SCRIPT_EXT, filename);
	s = GetRealFilename(buf, ScriptPath);
	g_free(buf);
    }
    else
	s = GetRealFilename(filename, ScriptPath);

    return s;
}

/* ScriptGenListFromFile() {{{1 */
static GList *
ScriptGenListFromFile(const char *filename)
{
    GList *list = NULL;
    char *buf;

    buf = ScriptGetFullPath(filename);
    if (buf && *buf)
	list = ScriptMakeListFromFile(buf);
    else
	StatusShowMessage(_("Cannot find '%s' with default extention/path"),
			  filename);

    return list;
}

/* ScriptRunFromFile() {{{1 */
void
ScriptRunFromFile(const char *filename)
{
    GList *list;

    if ((list = ScriptGenListFromFile(filename)) != NULL)
    {
	ScriptRunning = TRUE;
	ScriptInitCurrScriptItem(list);
	CurrScriptFile = g_strdup(filename);
	ScriptResult = OK;
	ScriptRunAnimStart();
	ScriptInstallNextItem();
    }
    else
	ScriptRunning = FALSE;
}

/* ScriptMakeListFromString() {{{1 */
static GList *
ScriptMakeListFromString(const char *inputStr)
{
    int n, result;
    GList *list;
    ScriptType *script;
    char *param[MAX_SCRIPT_PARAMS];
    char *s, *beginSentence, *freePtr;

    list = NULL;
    s = freePtr = g_strdup(inputStr);

    OptPrintf(OPRT_SCRIPT, "%s: s=%s\n", __FUNCTION__, s);
    while (*s)
    {
	beginSentence = s;	/*  ó  */

	/* 峡 ã´. */
	for (; *s; s++)
	{
	    /* double quote ȿ   𸣴 峡 ǹϴ ';'
	     * skipѴ.
	     */
	    if (*s == '"' || *s == '\'')
	    {
		char c = *s;
		do
		{
		    ++s;
		    if (*s == c)
			break;
		}
		while (*s);
	    }
	    if (*s == ';')
		break;
	}

	/* ';' ƴ϶ *s ݵ NUL ̾ Ѵ. */
	if (*s == ';')
	    *s++ = '\0';
	else if (*s)
	    g_warning("%s: parser error", __FUNCTION__);

	if ((n = ParseLine(beginSentence, MAX_SCRIPT_PARAMS, param)) < 1)
	    continue;

	OptPrintf(OPRT_SCRIPT, "%s: n=%d, param[0]=%s, *s=%d\n", __FUNCTION__,
		  n, param[0], *s);

	script = ScriptCreateItem(param, n - 1, 0, ScriptMainTable, &result);
	if (script == NULL)
	{
	    if (result == OK)
		/* just ignore this line! */
		continue;

	    break;
	}

	if (script->type == SCR_INCLUDE)
	    list = ScriptInclude(list, script);
	else
	    list = g_list_append(list, script);
    }
    g_free(freePtr);
    return list;
}

/* ScriptRunFromString() {{{1 */
void
ScriptRunFromString(const char *str)
{
    GList *list;

    if (str)
    {
	if ((list = ScriptMakeListFromString(str)) != NULL)
	{
	    ScriptRunning = TRUE;
	    ScriptInitCurrScriptItem(list);
	    CurrScriptFile = g_strdup("string");
	    ScriptResult = OK;
	    ScriptRunAnimStart();
	    ScriptInstallNextItem();
	}
	else
	    StatusShowMessage(_("Fail to run '%s'."), str);
    }
}

/* IsSystemScript() {{{1 */
/* file login/logout/startup script̸ TRUE return */
static gboolean
IsSystemScript(const char *filename)
{
    GSList *slist;
    ModemInfoType *mi;
    TelnetInfoType *ti;

    for (slist = ModemInfoList; slist; slist = slist->next)
    {
	mi = slist->data;
	if (mi->loginScript && !strcmp(mi->loginScript, filename))
	    return TRUE;
	if (mi->logoutScript && !strcmp(mi->logoutScript, filename))
	    return TRUE;
    }
    for (slist = TelnetInfoList; slist; slist = slist->next)
    {
	ti = slist->data;
	if (ti->loginScript && !strcmp(ti->loginScript, filename))
	    return TRUE;
	if (ti->logoutScript && !strcmp(ti->logoutScript, filename))
	    return TRUE;
    }
    if (!strcmp(filename, DEFAULT_STARTUP_SCRIPT))
	return TRUE;

    return FALSE;
}

/* CreateScriptListFromPath() {{{1 */
static GList *
CreateScriptListFromPath(char *fullpath, GList *list)
{
    register guint i;
    glob_t gl;
    char *filename, *p;

    if (glob(fullpath, 0, NULL, &gl) == 0)
    {
	for (i = 0; i < gl.gl_pathc; i++)
	{
	    filename = g_strdup(g_basename(gl.gl_pathv[i]));
	    /* system script combo box    Ѵ.
	     *     script  ȭڿ ...
	     *
	     * NOTE:   ణ  ִ , װ full path ƴ
	     * ϸ  Ѵٴ ̴.
	     */
	    if (IsSystemScript(filename))
		continue;
	    if ((p = strrchr(filename, '.')) != NULL)
		*p = '\0';
	    list = g_list_append(list, filename);
	}
	globfree(&gl);
    }

    return list;
}

/* CreateScriptFileList() {{{1 */
/* ⿡ GList  ϴ , combo GList ؾ Ǳ .
 */
GtkWidget *
CreateScriptFileList(GtkWidget *w, GtkSignalFunc runCB)
{
    GtkWidget *cb = NULL;
    GList *list = NULL;
    char buf[MAX_PATH];

    /* current directory ִ script file */
    g_snprintf(buf, sizeof(buf), "./*" DEFAULT_SCRIPT_EXT);
    list = CreateScriptListFromPath(buf, list);

    /* ScriptPath ִ script file */
    g_snprintf(buf, sizeof(buf), "%s/*" DEFAULT_SCRIPT_EXT, ScriptPath);
    list = CreateScriptListFromPath(buf, list);

    if (list)
    {
	guint width;

	cb = gtk_combo_new();
	gtk_combo_set_popdown_strings(GTK_COMBO(cb), list);
	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(cb)->entry),
			   (gchar *) list->data);
	gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(cb)->entry), FALSE);
	ScriptBox = GTK_WIDGET(GTK_COMBO(cb)->entry);
	gtk_signal_connect(GTK_OBJECT(GTK_COMBO(cb)->list),
			   "button_press_event", runCB, ScriptBox);

	width = gdk_string_width(gtk_style_get_font(ScriptBox->style), "A") * 12;
	gtk_widget_set_usize(ScriptBox, width, -1);
    }

    return cb;

    w = NULL;	/* to make gcc happy */
}
/* }}} */

/****************************************************************************
			       A N I M A T I O N
****************************************************************************/
/* ScriptAnimType {{{1 */
#define SA_LARGE_PIXMAP_SIZE	32
#define SA_SMALL_PIXMAP_SIZE	16
typedef struct {
    GtkWidget *w;
    GdkPixmap *pixmap;
    GdkBitmap *mask;
    char *filename;
    int currFrame;
    int frameNum;
    int size;
    guint tag;
} ScriptAnimType;

static ScriptAnimType *ScriptAnim = NULL;

/* ScriptRunNextFrame() {{{1 */
static gint
ScriptRunNextFrame(gpointer data)
{
    ScriptAnimType *sa = data;

    sa->currFrame = (sa->currFrame + 1) % sa->frameNum;
    if (!sa->currFrame)
	sa->currFrame = 1;

/* FIXME: clearŰ  ãƾ enable  ִ. */
/* #define SA_TRANSPARENT_ENABLE 1 */

#ifdef SA_TRANSPARENT_ENABLE
    /* WHY clear doesn't work? */
    /* gdk_window_clear(sa->w->window); */
    gdk_window_clear_area(sa->w->window, sa->w->allocation.x,
			  sa->w->allocation.y,
			  sa->w->allocation.width, sa->w->allocation.height);
    {
	static int first = TRUE;
	if (first)
	{
	    first = FALSE;
	    gdk_window_shape_combine_mask(sa->w->window, sa->mask, 0, 0);
	}
    }
#endif
    gtk_widget_draw(sa->w, NULL);
    return TRUE;
}

/* ScriptRunExpose() {{{1 */
static gint
ScriptRunExpose(GtkWidget *w, GdkEventExpose *ev, gpointer data)
{
    ScriptAnimType *sa = data;

    gdk_draw_pixmap(sa->w->window, sa->w->style->black_gc, sa->pixmap,
		    sa->currFrame * sa->size, 0, 0, 0, sa->size, sa->size);
    return TRUE;

    w = NULL;
    ev = NULL;
}

/* ScriptRunAnimInit() {{{1 */
GtkWidget *
ScriptRunAnimInit(GtkSignalFunc runCB)
{
#if GTK_MAJOR_VERSION < 2
/* imlib {{{ */
    int width, height;
    char *filename;
    GdkImlibImage *im;
    ScriptAnimType *sa;

    if (!ScriptAnim && runCB)
    {
#ifdef SA_TRANSPARENT_ENABLE
	filename = GetRealFilename("tux-anim.xpm", PKGDATADIR "/pixmap");
#else
	filename = GetRealFilename(ScriptAnimFile, PKGDATADIR "/pixmap");
#endif
	if (filename == NULL)
	    return NULL;

	ScriptAnim = g_new(ScriptAnimType, 1);

	if (!strcasecmp(ToolBarType, "BOTH"))
	    ScriptAnim->size = SA_LARGE_PIXMAP_SIZE;
	else
	    ScriptAnim->size = SA_SMALL_PIXMAP_SIZE;

	ScriptAnim->filename = filename;
	ScriptAnim->pixmap = NULL;
	ScriptAnim->mask = NULL;

	gtk_widget_push_visual(gdk_imlib_get_visual());
	gtk_widget_push_colormap(gdk_imlib_get_colormap());
	ScriptAnim->w = gtk_drawing_area_new();
	gtk_widget_pop_colormap();
	gtk_widget_pop_visual();
	gtk_drawing_area_size(GTK_DRAWING_AREA(ScriptAnim->w),
			      ScriptAnim->size, ScriptAnim->size);
	gtk_signal_connect(GTK_OBJECT(ScriptAnim->w), "expose_event",
			   (GtkSignalFunc) ScriptRunExpose, ScriptAnim);
	gtk_signal_connect(GTK_OBJECT(ScriptAnim->w), "button_press_event",
			   runCB, ScriptAnim);
	gtk_widget_set_events(GTK_WIDGET(ScriptAnim->w),
			      GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK);
    }

    if (ScriptAnim && ScriptAnim->filename)
    {
	sa = ScriptAnim;
	im = gdk_imlib_load_image(sa->filename);

	height = sa->size;
	width = im->rgb_width * ((double) height / im->rgb_height);
	gdk_imlib_render(im, width, height);

	if (sa->pixmap)
	    gdk_imlib_free_pixmap(sa->pixmap);
	if (sa->mask)
	    gdk_imlib_free_bitmap(sa->mask);
	sa->pixmap = gdk_imlib_move_image(im);
	sa->mask = gdk_imlib_move_mask(im);
	gdk_imlib_destroy_image(im);

	sa->frameNum = width / height;
	sa->currFrame = 0;

	return sa->w;
    }
/* }}} */
#else
    int width, height;
    char *filename;
    GdkPixbuf *pb;
    ScriptAnimType *sa;

    if (!ScriptAnim && runCB)
    {
	filename = GetRealFilename(ScriptAnimFile, PKGDATADIR "/pixmap");
	if (filename == NULL)
	    return NULL;

	ScriptAnim = g_new(ScriptAnimType, 1);
	sa = ScriptAnim;

	if (!strcasecmp(ToolBarType, "BOTH"))
	    sa->size = SA_LARGE_PIXMAP_SIZE;
	else
	    sa->size = SA_SMALL_PIXMAP_SIZE;

	sa->filename = filename;
	sa->pixmap = NULL;
	sa->mask = NULL;

	pb = gdk_pixbuf_new_from_file(filename, NULL);
	if (pb == NULL)
	    return NULL;

	width = gdk_pixbuf_get_width(pb);
	height = gdk_pixbuf_get_height(pb);
	if (height != sa->size)
	{
	    GdkPixbuf *pbt;
	    width = width * ((double)sa->size / height);
	    height = sa->size;
	    pbt = gdk_pixbuf_scale_simple(pb, width, height,
					  GDK_INTERP_NEAREST);
	    gdk_pixbuf_unref(pb);
	    pb = pbt;
	}

	sa->frameNum = width / height;
	sa->currFrame = 0;

	gdk_pixbuf_render_pixmap_and_mask(pb, &sa->pixmap, &sa->mask, 128);
	gdk_pixbuf_unref(pb);

	sa->w = gtk_drawing_area_new();
	gtk_widget_ref(sa->w);
	gtk_drawing_area_size(GTK_DRAWING_AREA(sa->w), sa->size, sa->size);
	gtk_signal_connect(GTK_OBJECT(sa->w), "expose_event",
			   (GtkSignalFunc) ScriptRunExpose, sa);
	gtk_signal_connect(GTK_OBJECT(sa->w), "button_press_event",
			   runCB, sa);
	gtk_widget_set_events(GTK_WIDGET(sa->w),
			      GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK);
	gtk_widget_show(sa->w);
    }

    if (ScriptAnim && ScriptAnim->filename)
	return ScriptAnim->w;

#endif
    return NULL;
}

/* ScriptRunAnimDestroy() {{{1 */
void
ScriptRunAnimDestroy(void)
{
    if (ScriptAnim)
    {
	gtk_widget_destroy(ScriptAnim->w);
	g_free(ScriptAnim->filename);
	g_free(ScriptAnim);
	ScriptAnim = NULL;
    }
}

/* ScriptRunSizeChanged() {{{1 */
void
ScriptRunSizeChanged(gboolean large)
{
    ScriptAnimType *sa = ScriptAnim;
    if (sa)
    {
	if (large)
	    sa->size = SA_LARGE_PIXMAP_SIZE;
	else
	    sa->size = SA_SMALL_PIXMAP_SIZE;
	gtk_drawing_area_size(GTK_DRAWING_AREA(sa->w), sa->size, sa->size);
	gtk_widget_set_usize(GTK_WIDGET(sa->w), sa->size, sa->size);
	ScriptRunAnimInit(NULL);
    }
}

/* ScriptRunAnimStart() {{{1 */
static void
ScriptRunAnimStart(void)
{
    if (ScriptAnim && ScriptAnim->tag == 0)
	ScriptAnim->tag = gtk_timeout_add(300, ScriptRunNextFrame, ScriptAnim);
}

/* ScriptRunAnimStop() {{{1 */
static void
ScriptRunAnimStop(void)
{
    if (ScriptAnim && ScriptAnim->tag != 0)
    {
	gtk_timeout_remove(ScriptAnim->tag);
	ScriptAnim->tag = 0;
	ScriptAnim->currFrame = 0;
	gtk_widget_draw(ScriptAnim->w, NULL);
    }
}
