/*
 * Status bar
 *
 * Copyright (C) 2000-2002, Nam SungHyun and various contributors
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#include "config.h"

#include "pcMain.h"
#include "pcTerm.h"	/* Term */

static GtkWidget *StatusLabel = NULL;
static GtkWidget *TimeLabel = NULL;
static GMutex *progMutex = NULL;

static gint
ProgressBarUpdate(gpointer data)
{
    static gfloat oldVal = 0;
    gfloat f = TRxPercent;

    if (f != oldVal)
    {
	if (f < 0.5)
	    f = 0.5;
	else if (f > 100.0)
	    f = 100.0;
	/* update bar */
	g_mutex_lock(progMutex);
	gtk_progress_set_value(GTK_PROGRESS(data), f);
	g_mutex_unlock(progMutex);
    }

    oldVal = f;

    return TRUE;
}

void ProgressBarControl(gboolean start)
{
    GtkAdjustment *adj;
    static GtkWidget *align, *pbar = NULL;
    static guint ProgressTag;

    if (start)
    {
	if (pbar == NULL)
	{
	    guint height;

	    /* ʱⰪ 0.0̸ ʿ -1% ǥõȴ. */
	    TRxPercent = 0.5;
	    progMutex = g_mutex_new();
	    g_mutex_lock(progMutex);

	    /* progress bar */
	    align = gtk_alignment_new(0.5, 0.5, 0, 0);
	    gtk_box_pack_end(GTK_BOX(StatBox), align, FALSE, FALSE, 5);
	    gtk_container_set_border_width(GTK_CONTAINER(align), 0);
	    adj = (GtkAdjustment *) gtk_adjustment_new(0, 1, 100, 0, 0, 0);
	    pbar = gtk_progress_bar_new_with_adjustment(adj);
	    height = gtk_style_get_font(StatusLabel->style)->ascent
		+ gtk_style_get_font(StatusLabel->style)->descent;
	    gtk_widget_set_usize(pbar, 100, height + 6);
	    gtk_progress_set_format_string(GTK_PROGRESS(pbar), "%p%%");
	    gtk_progress_bar_set_orientation(GTK_PROGRESS_BAR(pbar),
					     GTK_PROGRESS_LEFT_TO_RIGHT);
	    gtk_progress_bar_set_bar_style(GTK_PROGRESS_BAR(pbar),
					   GTK_PROGRESS_CONTINUOUS);
	    gtk_container_add(GTK_CONTAINER(align), pbar);
	    gtk_widget_show(pbar);
	    gtk_widget_show(align);

	    gtk_progress_set_show_text(GTK_PROGRESS(pbar), TRUE);
	    gtk_progress_set_value(GTK_PROGRESS(pbar), TRxPercent);
	    gtk_progress_set_text_alignment(GTK_PROGRESS(pbar), 0.5, 1.0);

	    ProgressTag = gtk_idle_add(ProgressBarUpdate, pbar);
	    g_mutex_unlock(progMutex);
	}
    }
    else
    {
	if (pbar)
	{
	    g_mutex_lock(progMutex);
	    gtk_idle_remove(ProgressTag);
	    gtk_widget_destroy(pbar);
	    gtk_widget_destroy(align);
	    pbar = NULL;
	    g_mutex_unlock(progMutex);
	    g_mutex_free(progMutex);
	}
    }
}

GtkWidget *StatusAddFrameWithLabel(const char *str)
{
    GtkWidget *frame, *label;

    frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_OUT);
    gtk_container_set_border_width(GTK_CONTAINER(frame), 0);
    gtk_box_pack_end(GTK_BOX(StatBox), frame, FALSE, FALSE, 0);
    gtk_widget_show(frame);

    label = gtk_label_new(str);
    gtk_container_add(GTK_CONTAINER(frame), label);
    gtk_widget_show(label);

    return frame;
}

void StatusRemoveFrame(GtkWidget *frame)
{
    gtk_widget_destroy(frame);
}

void CreateStatusBar(GtkWidget *w)
{
    GtkWidget *frame;

    StatusLabel = gtk_label_new(_("Welcome!!!"));
    gtk_box_pack_start(GTK_BOX(w), StatusLabel, FALSE, FALSE, 5);
    gtk_misc_set_alignment(GTK_MISC(StatusLabel), 0.0, 0.5);
    gtk_widget_show(StatusLabel);

    /* ð ̺ */
    frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_OUT);
    gtk_container_set_border_width(GTK_CONTAINER(frame), 0);
    gtk_box_pack_end(GTK_BOX(w), frame, FALSE, FALSE, 0);
    gtk_widget_show(frame);

    TimeLabel = gtk_label_new(" 00:00:00 ");
    gtk_container_add(GTK_CONTAINER(frame), TimeLabel);
    gtk_widget_show(TimeLabel);
}

void StatusShowMessage(const gchar *format, ...)
{
    va_list args;
    gchar *buf, *p;

    va_start(args, format);
    buf = g_strdup_vprintf(format, args);
    va_end(args);

    /* FIXME: status area ̸ ͼ ó */
    if (strlen(buf) > 60)
	buf[60] = '\0';
    /* change '\n' to space */
    for (p = buf; (p = strchr(p, '\n')) != NULL; p++)
	*p = ' ';

    gtk_label_set_text(GTK_LABEL(StatusLabel), buf);
    g_free(buf);
}

void StatusShowTime(gint32 secs)
{
    gchar buf[20];
    gint hour;

    if (GTK_WIDGET_HAS_FOCUS(Term))
    {
	hour = (gint) secs / 3600;
	secs %= 3600UL;
	g_snprintf(buf, sizeof(buf), " %02d:%02d:%02d ", hour,
		   (gint) (secs / 60), (gint) (secs % 60));
	gtk_label_set_text(GTK_LABEL(TimeLabel), buf);
    }
}

GtkWidget *DialogShowMessage(GtkWidget *parent, const char *title,
			     const char *labelName, const char *msg)
{
    GtkWidget *w;
    GtkWidget *vbox, *label, *sep, *button;

    g_return_val_if_fail(msg != NULL, NULL);

    w = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    if (title)
	gtk_window_set_title(GTK_WINDOW(w), title);
    else
	gtk_window_set_title(GTK_WINDOW(w), PACKAGE " - " VERSION);
    if (parent)
	gtk_window_set_transient_for(GTK_WINDOW(w), GTK_WINDOW(parent));
    gtk_container_set_border_width(GTK_CONTAINER(w), 4);
    gtk_window_set_position(GTK_WINDOW(w), GTK_WIN_POS_MOUSE);

    vbox = gtk_vbox_new(FALSE, 2);
    gtk_container_add(GTK_CONTAINER(w), vbox);

    label = gtk_label_new(msg);
    gtk_container_add(GTK_CONTAINER(vbox), label);

    sep = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(vbox), sep, FALSE, FALSE, 1);

    button = gtk_button_new_with_label(labelName);
    gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 1);
    gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
			      GTK_SIGNAL_FUNC(gtk_widget_destroy),
			      GTK_OBJECT(w));

    gtk_widget_show_all(w);
    return w;
}
