/*
 * telnet
 *
 * Copyright (C) 2000-2002, Nam SungHyun and various contributors
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#include "config.h"
#include <signal.h>
#include <sys/wait.h>	/* waitpid() */

#include "pcMain.h"
#include "pcTerm.h"

/* Global variables/functions {{{1 */

char *TelnetCommand;

/*  Ͽ о telnet info  */
GSList *TelnetInfoList = NULL;

/*   telnet info. */
TelnetInfoType *TelnetInfo = NULL;

/* Local function prototypes {{{1 */

static void TelnetInfoFree(void);

/* Local variables {{{1 */

/*  telnet ӵ ð */
static guint32 TelnetConnectTime = 0UL;

/* telnet  帥 ʴ tick counter */
static gint32 TelnetTimeTick;

/* telnet process id */
static pid_t TelnetPid = -1;

static gboolean UseSSH = FALSE;

static TelnetInfoType DfltTelnetInfo = { NULL, NULL, NULL, NULL, NULL };

/* TelnetTimeReport() {{{1 */
static int
TelnetTimeReport(void)
{
    register int newTick;
    guint32 currTime;
    TermType *term;

    currTime = gdk_time_get();
    if (TelnetConnectTime)
    {
	newTick = (currTime - TelnetConnectTime) / 1000;
	if (newTick != TelnetTimeTick)
	{
	    TelnetTimeTick = newTick;
	    StatusShowTime(newTick);
	}
    }
    if (ReadFD > 0)
    {
	if (IdleGuardString && *IdleGuardString)
	    if ((LastOutTime + IdleGuardInterval * 1000) < currTime)
		Term->remoteWrite(IdleGuardString, strlen(IdleGuardString));

	if (AR_CheckFlag)
	{
	    AR_CheckFlag = FALSE;
	    term = Term;
	    AutoResponseCheckRun(term->buf
				 + term->col * TERM_LINENUM(term->currY),
				 term->currX);
	}

	return TRUE;
    }
    else
    {
	StatusShowMessage(_("Disconnected from %s."), TelnetInfo->name);
	SoundPlay(SoundDisconnectFile);

	/* logging */
	if (LogFile)
	    CloseLogFile(TelnetTimeTick);

	PC_State = PC_IDLE_STATE;
	PC_ButtonControl(TB_IDLE_STATE);

	if (TelnetInfo->logoutScript)
	    ScriptRunFromFile(TelnetInfo->logoutScript);

	TelnetInfoFree();

	return FALSE;
    }
}

/* TelnetWrite() {{{1 */
static int
TelnetWrite(const guchar *buf, int len)
{
    register int n;

    if (!buf || !*buf || len <= 0)
	return 0;

    LastOutTime = gdk_time_get();

    if (DoLocalEcho && Term)
	Term->localWrite(Term, buf, len);

    if ((n = write(ReadFD, buf, len)) != len)
	g_warning("%s failure: %d, %d", __FUNCTION__, len, n);

    ControlLabelLineStsToggle('t');

    return len;
}

/* TelnetConnectionCheck() {{{1 */
static int
TelnetConnectionCheck(char *buf, int size)
{
    register int i;
    GSList *slist;
    static const char tcpConnectStr[] = "Escape";
    static const char *p = tcpConnectStr;

    for (i = 0; i < size; i++)
    {
	if (UseSSH == FALSE && buf[i] != *p)
	    p = tcpConnectStr;
	else if (*++p == '\0' || UseSSH == TRUE)
	{
	    /* connected */

	    /* logging */
	    if (LogFile)
		OpenLogFile(TelnetInfo->name, TelnetInfo->host);

	    InputFilterList = g_slist_remove(InputFilterList,
					     TelnetConnectionCheck);

	    TelnetTimeTick = -1;
	    StatusShowMessage(_("Connected to %s."), TelnetInfo->name);
	    SoundPlay(SoundConnectFile);
	    SetWindowTitle(TelnetInfo->name);

	    for (slist = ConnectFuncList; slist;)
	    {
		void (*func) (void) = slist->data;

		/* list remove Ͼ   list  
		 * ϴ. func ڱ ڽ remove  ֱ ̴.
		 */
		slist = slist->next;

		(*func) ();
	    }
	    gtk_timeout_add(DEFAULT_CONNECT_CHK_INTERVAL,
			    (GtkFunction) TelnetTimeReport, NULL);
	    TelnetConnectTime = gdk_time_get();

	    if (UseSSH)
		Term->localWrite(Term, buf, size);
	    break;
	}
    }
    return 0;
}

/* TelnetDone() {{{1 */
void
TelnetDone(void)
{
    GSList *slist;

    InputFilterList = g_slist_remove(InputFilterList, TelnetConnectionCheck);

    if (TelnetPid != -1)
    {
	/* kill the old telnet process */
	kill(TelnetPid, SIGKILL);
	waitpid(TelnetPid, NULL, 0);
	TelnetPid = -1;
    }

    if (ReadFD != -1)
    {
	close(ReadFD);
	ReadFD = -1;
    }

    if (TermReadInputTag > 0)
    {
	gdk_input_remove(TermReadInputTag);
	TermReadInputTag = 0;
    }
    Term->remoteWrite = DummyFromTerm;

    if (ScriptRunning)
	ScriptCancel();
    if (LogFile)
	CloseLogFile(0);	/* safety */

    for (slist = DisconnectFuncList; slist;)
    {
	void (*func) (void) = slist->data;

	/* list remove Ͼ   list   ϴ.
	 * func ڱ ڽ remove  ֱ ̴.
	 */
	slist = slist->next;

	(*func)();
    }

    SetWindowTitle(NULL);

    if (PC_State != PC_IDLE_STATE)
    {
	StatusShowMessage(_("Telnet connection cancelled."));
	PC_State = PC_IDLE_STATE;
	PC_ButtonControl(TB_IDLE_STATE);
    }

    LastOutTime += 800000000UL;
    TelnetConnectTime = 0;
}

/* TelnetOpen() {{{1 */
static int
TelnetOpen(char *host, char *port, int col, int row)
{
    int result;
    pid_t pid;
    int status;
    char *telnetFullCmd;

    if (port && *port && *port != '0')
	telnetFullCmd = g_strconcat(TelnetCommand, " ", host, " ", port, NULL);
    else
	telnetFullCmd = g_strconcat(TelnetCommand, " ", host, NULL);

    UseSSH = FALSE;
    if (!strncmp(TelnetCommand, "ssh", 3))
	UseSSH = TRUE;

    result = PTY_Open(telnetFullCmd, &TelnetPid, col, row);
    g_free(telnetFullCmd);

    pid = waitpid(TelnetPid, &status, WUNTRACED | WNOHANG);
    if (pid == TelnetPid && pid >= 0)
    {
	TelnetPid = -1;
	result = -1;
    }

    if (result > 0)
	AddChildSignalHandler(TelnetPid, TelnetDone);

    return result;
}

/* TelnetConnect() {{{1 */
static void
TelnetConnect(char *host, char *port)
{
    /* no baudrate */
    ControlBaudrateChange(0);

    /*  ȣƮ  Ư ȯ漳   Ƿ login script
     *  Ѵ.
     */
    if (TelnetInfo->loginScript)
    {
	ScriptRunFromFile(TelnetInfo->loginScript);
	PC_Sleep(100); /* script ǵ */
    }

    StatusShowMessage(_("Trying to connect to %s..."), host);

    PC_State = PC_RUN_STATE;
    PC_ButtonControl(TB_CONNECT_STATE);

    g_assert(Term != NULL);

    ReadFD = TelnetOpen(host, port, Term->col, Term->row);

    if (ReadFD >= 0)
    {
	TermReadInputTag = gdk_input_add(ReadFD, GDK_INPUT_READ,
					 TermInputHandler, NULL);
	InputFilterList =
	    g_slist_prepend(InputFilterList, TelnetConnectionCheck);
	Term->remoteWrite = TelnetWrite;
    }
    else
    {
	ScriptCancel();
	g_warning("Cannot open telnet. What's wrong?");
    }
}

/* TelnetInfoFree() {{{1 */
static void
TelnetInfoFree(void)
{
    if (TelnetInfo == &DfltTelnetInfo)
    {
	if (TelnetInfo->host)
	{
	    g_free(TelnetInfo->host);
	    TelnetInfo->host = NULL;
	}
	if (TelnetInfo->name)
	{
	    g_free(TelnetInfo->name);
	    TelnetInfo->name = NULL;
	}
    }
    TelnetInfo = NULL;
}

/* TelnetConnectByHostname() {{{1 */
gint
TelnetConnectByHostname(const char *host)
{
    GSList *slist;
    TelnetInfoType *ti = NULL;

    if (TelnetPid != -1)
    {
	StatusShowMessage(_("Currently connected. Close first..."));
	return FALSE;
    }

    g_return_val_if_fail(Term != NULL, -1);
    g_return_val_if_fail(host != NULL, -1);
    g_return_val_if_fail(*host != 0, -1);

    Term->remoteWrite = DummyFromTerm;

    UseModem = FALSE;

    for (slist = TelnetInfoList; slist; slist = slist->next)
    {
	ti = slist->data;
	if (!strcmp(ti->host, host) || !strcmp(ti->name, host))
	    break;
    }
    if (!slist)
    {
	ti = &DfltTelnetInfo;

	if (ti->host)
	    g_free(ti->host);
	if (ti->name)
	    g_free(ti->name);

	ti->host = g_strdup(host);
	ti->name = g_strdup(host);
	ti->port = NULL;
	ti->loginScript = NULL;
	ti->logoutScript = NULL;
    }
    TelnetInfo = ti;
    TelnetConnect(ti->host, ti->port);
    return FALSE;
}

/* TelnetExit() {{{1 */
void
TelnetExit(void)
{
    GSList *l;
    TelnetInfoType *ti;

    for (l = TelnetInfoList; l; l = l->next)
    {
	if ((ti = (TelnetInfoType *) l->data) == NULL)
	    continue;

	if (ti->name)
	    g_free(ti->name);
	if (ti->host)
	    g_free(ti->host);
	if (ti->port)
	    g_free(ti->port);
	if (ti->loginScript)
	    g_free(ti->loginScript);
	if (ti->logoutScript)
	    g_free(ti->logoutScript);
	g_free(ti);
    }
    if (TelnetInfoList)
    {
	g_slist_free(TelnetInfoList);
	TelnetInfoList = NULL;
    }

    TelnetInfoFree();
}

